#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import netius.common
import netius.clients

from . import http
from . import http2

BUFFER_RATIO = 1.5
""" The ratio for the calculus of the internal socket
buffer size from the maximum pending buffer size """

MIN_RATIO = 0.8
""" The ration for the calculus of the minimum pending
value this is going to be used to re-enable the operation
and start the filling of the buffer again """

MAX_PENDING = netius.CHUNK_SIZE * 8
""" The size in bytes considered to be the maximum
allowed in the sending buffer, this maximum value
avoids the starvation of the producer to consumer
relation that could cause memory problems """

class ProxyConnection(http2.HTTP2Connection):

    def open(self, *args, **kwargs):
        http2.HTTP2Connection.open(self, *args, **kwargs)
        if not self.is_open(): return
        self.parser.store = False
        self.parser.bind("on_headers", self.on_headers)
        self.parser.bind("on_partial", self.on_partial)
        self.parser.bind("on_available", self.on_available)
        self.parser.bind("on_unavailable", self.on_unavailable)

    def resolve_encoding(self, parser):
        pass

    def set_h2(self):
        http2.HTTP2Connection.set_h2(self)
        self.parser.bind("on_headers", self.on_headers)
        self.parser.bind("on_partial", self.on_partial)
        self.parser.bind("on_available", self.on_available)
        self.parser.bind("on_unavailable", self.on_unavailable)

    def on_headers(self):
        self.owner.on_headers(self.connection_ctx, self.parser_ctx)

    def on_partial(self, data):
        self.owner.on_partial(self.connection_ctx, self.parser_ctx, data)

    def on_available(self):
        self.owner.on_available(self.connection_ctx, self.parser_ctx)

    def on_unavailable(self):
        self.owner.on_unavailable(self.connection_ctx, self.parser_ctx)

class ProxyServer(http2.HTTP2Server):

    def __init__(
        self,
        dynamic = True,
        throttle = True,
        trust_origin = False,
        max_pending = MAX_PENDING,
        *args,
        **kwargs
    ):
        http2.HTTP2Server.__init__(
            self,
            receive_buffer_c = int(max_pending * BUFFER_RATIO),
            send_buffer_c = int(max_pending * BUFFER_RATIO),
            *args,
            **kwargs
        )
        self.dynamic = dynamic
        self.throttle = throttle
        self.trust_origin = trust_origin
        self.max_pending = max_pending
        self.min_pending = int(max_pending * MIN_RATIO)
        self.conn_map = {}

        self.http_client = netius.clients.HTTPClient(
            thread = False,
            auto_release = False,
            receive_buffer = max_pending,
            send_buffer = max_pending,
            *args,
            **kwargs
        )
        self.http_client.bind("headers", self._on_prx_headers)
        self.http_client.bind("message", self._on_prx_message)
        self.http_client.bind("partial", self._on_prx_partial)
        self.http_client.bind("connect", self._on_prx_connect)
        self.http_client.bind("acquire", self._on_prx_acquire)
        self.http_client.bind("close", self._on_prx_close)
        self.http_client.bind("error", self._on_prx_error)

        self.raw_client = netius.clients.RawClient(
            thread = False,
            receive_buffer = int(max_pending * BUFFER_RATIO),
            send_buffer = int(max_pending * BUFFER_RATIO),
            *args,
            **kwargs
        )
        self.raw_client.bind("connect", self._on_raw_connect)
        self.raw_client.bind("data", self._on_raw_data)
        self.raw_client.bind("close", self._on_raw_close)

        self.container = netius.Container(*args, **kwargs)
        self.container.add_base(self)
        self.container.add_base(self.http_client)
        self.container.add_base(self.raw_client)

    def start(self):
        # starts the container this should trigger the start of the
        # event loop in the container and the proper listening of all
        # the connections in the current environment
        self.container.start(self)

    def stop(self):
        # verifies if there's a container object currently defined in
        # the object and in case it does exist propagates the stop call
        # to the container so that the proper stop operation is performed
        if not self.container: return
        self.container.stop()

    def cleanup(self):
        http2.HTTP2Server.cleanup(self)

        # saves the container reference so that it may be used latter
        # and then unsets it under the current instance
        container = self.container
        self.container = None

        # verifies if the container is valid and if that's not the case
        # returns the control flow immediately (as expected)
        if not container: return

        # runs the cleanup operation on the cleanup, this should properly
        # propagate the operation to the owner container (as expected)
        container.cleanup()

        # unsets the references to the inner clients that compose the
        # proxy server, avoids possible memory leaks
        self.http_client = None
        self.raw_client = None

    def info_dict(self, full = False):
        info = http2.HTTP2Server.info_dict(self, full = full)
        info.update(
            dynamic = self.dynamic,
            throttle = self.throttle,
            max_pending = self.max_pending,
            min_pending = self.min_pending,
            http_client = self.http_client.info_dict(full = full),
            raw_client = self.raw_client.info_dict(full = full)
        )
        return info

    def connections_dict(self, full = False, parent = False):
        if parent: return http2.HTTP2Server.connections_dict(self, full = full)
        return self.container.connections_dict(full = full)

    def connection_dict(self, id, full = False):
        return self.container.connection_dict(id, full = full)

    def on_data(self, connection, data):
        netius.StreamServer.on_data(self, connection, data)

        # tries to retrieve the reference to the tunnel connection
        # currently set in the connection in case it does not exists
        # (initial handshake or http client proxy) runs the parse
        # step on the data and then returns immediately
        tunnel_c = hasattr(connection, "tunnel_c") and connection.tunnel_c
        if not tunnel_c: connection.parse(data); return

        # verifies that the current size of the pending buffer is greater
        # than the maximum size for the pending buffer the read operations
        # if that the case the read operations must be disabled
        should_throttle = self.throttle and connection.is_throttleable()
        should_disable = should_throttle and tunnel_c.is_exhausted()
        if should_disable: connection.disable_read()

        # performs the sending operation on the data but uses the throttle
        # callback so that the connection read operations may be resumed if
        # the buffer has reached certain (minimum) levels
        tunnel_c.send(data, callback = self._throttle)

    def on_connection_d(self, connection):
        http2.HTTP2Server.on_connection_d(self, connection)

        tunnel_c = hasattr(connection, "tunnel_c") and connection.tunnel_c
        proxy_c = hasattr(connection, "proxy_c") and connection.proxy_c

        if tunnel_c: tunnel_c.close()
        if proxy_c: proxy_c.close()

        setattr(connection, "tunnel_c", None)
        setattr(connection, "proxy_c", None)

    def on_stream_d(self, stream):
        http2.HTTP2Server.on_stream_d(self, stream)

        tunnel_c = hasattr(stream, "tunnel_c") and stream.tunnel_c
        proxy_c = hasattr(stream, "proxy_c") and stream.proxy_c

        if tunnel_c: tunnel_c.close()
        if proxy_c: proxy_c.close()

        setattr(stream, "tunnel_c", None)
        setattr(stream, "proxy_c", None)

    def on_serve(self):
        http2.HTTP2Server.on_serve(self)
        if self.env: self.dynamic = self.get_env("DYNAMIC", self.dynamic, cast = bool)
        if self.env: self.throttle = self.get_env("THROTTLE", self.throttle, cast = bool)
        if self.env: self.trust_origin = self.get_env("TRUST_ORIGIN", self.trust_origin, cast = bool)
        if self.dynamic: self.info("Using dynamic encoding (no content re-encoding) in proxy ...")
        if self.throttle: self.info("Throttling connections in proxy ...")
        else: self.info("Not throttling connections in proxy ...")
        if self.trust_origin: self.info("Origin is considered \"trustable\" by proxy")

    def on_data_http(self, connection, parser):
        http2.HTTP2Server.on_data_http(self, connection, parser)

        if not hasattr(connection, "proxy_c"): return

        proxy_c = connection.proxy_c

        should_throttle = self.throttle and connection.is_throttleable()
        should_disable = should_throttle and proxy_c.is_exhausted()
        if should_disable: connection.disable_read()
        proxy_c.flush(force = True, callback = self._throttle)

    def on_headers(self, connection, parser):
        pass

    def on_partial(self, connection, parser, data):
        if not hasattr(connection, "proxy_c"): return

        proxy_c = connection.proxy_c

        should_throttle = self.throttle and connection.is_throttleable()
        should_disable = should_throttle and proxy_c.is_exhausted()
        if should_disable: connection.disable_read()
        proxy_c.send_base(data, force = True, callback = self._throttle)

    def on_available(self, connection, parser):
        proxy_c = connection.proxy_c
        if not proxy_c.renable == False: return
        if not connection.is_restored(): return
        proxy_c.enable_read()
        self.reads((proxy_c.socket,), state = False)

    def on_unavailable(self, connection, parser):
        proxy_c = connection.proxy_c
        if proxy_c.renable == False: return
        should_throttle = self.throttle and proxy_c.is_throttleable()
        should_disable = should_throttle and connection.is_exhausted()
        if not should_disable: return
        proxy_c.disable_read()

    def build_connection(self, socket, address, ssl = False):
        return ProxyConnection(
            owner = self,
            socket = socket,
            address = address,
            ssl = ssl,
            encoding = self.encoding,
            max_pending = self.max_pending,
            min_pending = self.min_pending
        )

    def _throttle(self, _connection):
        if not _connection.is_restored(): return
        connection = self.conn_map[_connection]
        if not connection.renable == False: return
        connection.enable_read()
        self.reads((connection.socket,), state = False)

    def _prx_close(self, connection):
        connection.close(flush = True)

    def _prx_keep(self, connection):
        pass

    def _prx_throttle(self, connection):
        if not connection.is_restored(): return

        proxy_c = hasattr(connection, "proxy_c") and connection.proxy_c
        if not proxy_c: return
        if not proxy_c.renable == False: return

        proxy_c.enable_read()
        self.http_client.reads((proxy_c.socket,), state = False)

    def _raw_throttle(self, connection):
        if not connection.is_restored(): return

        tunnel_c = hasattr(connection, "tunnel_c") and connection.tunnel_c
        if not tunnel_c: return
        if not tunnel_c.renable == False: return

        tunnel_c.enable_read()
        self.raw_client.reads((tunnel_c.socket,), state = False)

    def _on_prx_headers(self, client, parser, headers):
        # retrieves the owner of the parser as the client connection
        # and then retrieves all the other http specific values
        _connection = parser.owner
        code_s = parser.code_s
        status_s = parser.status_s
        version_s = parser.version_s

        # creates a new dictionary from the provided one, so that no overlap
        # in values occurs (would destroy the original data)
        headers = dict(headers)

        # resolves the client connection into the proper proxy connection
        # to be used to send the headers (and status line) to the client
        # and then retrieves the origin content and transfer encoding values
        # that are going to be used to determine some heuristics for the data
        connection = self.conn_map[_connection]
        content_encoding = headers.pop("content-encoding", None)
        transfer_encoding = headers.pop("transfer-encoding", None)

        # if either the proxy connection or the back-end one is compressed
        # the length values of the connection are considered unreliable and
        # some extra operation must be defined, note that in case the dynamic
        # (no re-encoding) support is enabled the length is always reliable
        unreliable_length = _connection.current > http.CHUNKED_ENCODING or\
            connection.current > http.CHUNKED_ENCODING or parser.content_l == -1
        unreliable_length &= not self.dynamic

        # in case the content length is unreliable some of the headers defined
        # must be removed so that no extra connection error occurs, as the size
        # of the content from one end point to the other may change
        if unreliable_length:
            if "content-length" in headers: del headers["content-length"]
            if "accept-ranges" in headers: del headers["accept-ranges"]

        # in case the length of the data is not reliable and the current connection
        # is plain encoded a proper set of operation must be properly handled including
        # the forcing of the chunked encoding or the connection drop at the end of the
        # message strategies
        if unreliable_length and connection.current < http.CHUNKED_ENCODING:
            if parser.version < netius.common.HTTP_11:
                connection.parser.keep_alive = False
            else:
                connection.set_encoding(http.CHUNKED_ENCODING)

        # tries to use the content encoding value to determine the minimum encoding
        # that allows the content encoding to be kept (compatibility support), this
        # heuristic is only applied in case the dynamic option is enabled
        content_encoding_c = http.ENCODING_MAP.get(content_encoding, connection.current)
        content_encoding_t = http.ENCODING_MAP.get(transfer_encoding, connection.current)
        target_encoding = max(content_encoding_c, content_encoding_t)
        if self.dynamic and target_encoding > connection.current:
            connection.set_encoding(target_encoding)

        # applies the headers meaning that the headers are going to be
        # processed so that they represent the proper proxy operation
        # that is going to be done with the passing of the data, note
        # that the connection request context is applied so that the
        # proper encoding is used in the connection header application
        with connection.ctx_request():
            self._apply_headers(connection.parser, connection, parser, headers)

        # runs the send headers operation that will start the transmission
        # of the headers for the requested proxy operation, the concrete
        # semantics of the transmission should dependent on the version of
        # the protocol that is going to be used in the transmission
        connection.send_header(
            headers = headers,
            version = version_s,
            code = int(code_s),
            code_s = status_s
        )

    def _on_prx_message(self, client, parser, message):
        # retrieves the back-end connection from the provided parser this
        # is going to be used for the reverse connection resolution process
        _connection = parser.owner

        # sets the current client connection as not waiting and then retrieves
        # the requester connection associated with the client (back-end)
        # connection in order to be used in the current processing
        _connection.waiting = False
        connection = self.conn_map[_connection]

        # creates the clojure function that will be used to close the
        # current client connection and that may or may not close the
        # corresponding back-end connection (as defined in specification)
        def close(connection): connection.close(flush = True)

        # verifies that the connection is meant to be kept alive, the
        # connection is meant to be kept alive when both the client and
        # the final (back-end) server respond with the keep alive flag
        keep_alive = parser.keep_alive and connection.parser.keep_alive

        # defines the proper callback function to be called at the end
        # of the flushing of the connection according to the result of
        # the keep alive evaluation (as defined in specification)
        if keep_alive: callback = None
        else: callback = close

        # runs the final flush operation in the connection making sure that
        # every data that is pending is properly flushed, this is especially
        # important for chunked or compressed connections
        connection.flush_s(callback = callback)

    def _on_prx_partial(self, client, parser, data):
        # retrieves the owner of the proxy parser as the proxy connection
        # and then uses the connection to decode the data to obtain the raw
        # value of it, this is required as gzip compression may exist
        _connection = parser.owner
        data = data if self.dynamic else _connection.raw_data(data)

        # retrieves the peer connection and tries to send the new data chunk
        # back to it using the currently defined encoding (as expected), note
        # that additional throttling operations may apply
        connection = self.conn_map[_connection]
        should_throttle = self.throttle and _connection.is_throttleable()
        should_disable = should_throttle and connection.is_exhausted()
        if should_disable: _connection.disable_read()
        connection.send_part(data, final = False, callback = self._prx_throttle)

    def _on_prx_connect(self, client, _connection):
        _connection.waiting = False

    def _on_prx_acquire(self, client, _connection):
        _connection.waiting = False

    def _on_prx_close(self, client, _connection):
        # retrieves the reference to the parent class value
        # so that it can be used for class level operations
        cls = self.__class__

        # retrieves the front-end connection associated with
        # the back-end to be used for the operations in case
        # no connection is retrieved returns the control flow
        # to the caller method immediately (nothing done)
        connection = self.conn_map.get(_connection, None)
        if not connection: return

        # in case the connection is under the waiting state
        # the forbidden response is set to the client otherwise
        # the front-end connection is closed immediately
        if _connection.waiting: connection.send_response(
            data = cls.build_data(
                "Forbidden",
                url = _connection.error_url if\
                    hasattr(_connection, "error_url") else None
            ),
            headers = dict(
                connection = "close"
            ),
            code = 403,
            code_s = "Forbidden",
            apply = True,
            callback = self._prx_close
        )
        else: connection.close(flush = True)

        # removes the waiting state from the connection and
        # the removes the back-end to front-end connection
        # relation for the current proxy connection
        _connection.waiting = False
        del self.conn_map[_connection]

    def _on_prx_error(self, client, _connection, error):
        # retrieves the reference to the parent class value
        # so that it can be used for class level operations
        cls = self.__class__

        # retrieves the front-end connection associated with
        # the proxy connection, this value is going to be
        # if sending the message to the final client
        connection = self.conn_map.get(_connection, None)
        if not connection: return

        # constructs the message string that is going to be
        # sent as part of the response from the proxy indicating
        # the unexpected error, then in case the connection is
        # still under the (initial) waiting state sends the same
        # message to the final client connection (indicating error)
        # note that the recovery from the error (disconnect should
        # be handled by the error manager, and that should imply
        # a closing operation on the original/proxy connection)
        error_m = str(error) or "Unknown proxy relay error"
        if _connection.waiting: connection.send_response(
            data = cls.build_text(error_m),
            headers = dict(
                connection = "close"
            ),
            code = 500,
            code_s = "Internal Error",
            apply = True
        )

        # sets the connection as not waiting, so that no more
        # messages are sent as part of the closing chain
        _connection.waiting = False

    def _on_raw_connect(self, client, _connection):
        connection = self.conn_map[_connection]
        connection.send_response(
            code = 200,
            code_s = "Connection established",
            apply = True
        )

    def _on_raw_data(self, client, _connection, data):
        connection = self.conn_map[_connection]
        should_throttle = self.throttle and _connection.is_throttleable()
        should_disable = should_throttle and connection.is_exhausted()
        if should_disable: _connection.disable_read()
        connection.send(data, callback = self._raw_throttle)

    def _on_raw_close(self, client, _connection):
        connection = self.conn_map[_connection]
        connection.close(flush = True)
        del self.conn_map[_connection]

    def _apply_headers(self, parser, connection, parser_prx, headers, upper = True):
        if upper: self._headers_upper(headers)
        self._apply_via(parser_prx, headers)
        self._apply_all(parser_prx, connection, headers, replace = True)

    def _apply_via(self, parser_prx, headers):
        # retrieves the various elements of the parser that are going
        # to be used for the creation of the via string value, and
        # processes some of them to take them into the normal form
        connection = parser_prx.owner
        version_s = parser_prx.version_s
        version_s = version_s.split("/", 1)[1]

        # unpacks the current connectiont's address so that the host
        # value is possible to be retrieved (as expected)
        host, _port = connection.address

        # retrieves the server value from the current headers, as it
        # is going to be used for the creation of the partial via
        # value (the technology part of the string)
        server = headers.get("Server", None)

        # creates the via string value taking into account if the server
        # part of the string exists or not (different template)
        if server: via_s = "%s %s (%s)" % (version_s, host, server)
        else: via_s = "%s %s" % (version_s, host)

        # tries to retrieve the current via string (may already exits)
        # and appends the created string to the base string or creates
        # a new one (as defined in the http specification)
        via = headers.get("Via", "")
        if via: via += ", "
        via += via_s
        headers["Via"] = via



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import struct

import netius.common
import netius.clients

GRANTED = 0x5a
REJECTED = 0x5b
FAILED_CLIENT = 0x5c
FAILED_AUTH = 0x5d

GRANTED_EXTRA = 0x00

BUFFER_RATIO = 1.5
""" The ratio for the calculus of the internal socket
buffer size from the maximum pending buffer size """

MIN_RATIO = 0.8
""" The ration for the calculus of the minimum pending
value this is going to be used to re-enable the operation
and start the filling of the buffer again """

MAX_PENDING = 65536
""" The size in bytes considered to be the maximum
allowed in the sending buffer, this maximum value
avoids the starvation of the producer to consumer
relation that could cause memory problems """

class SOCKSConnection(netius.Connection):

    def __init__(self, *args, **kwargs):
        netius.Connection.__init__(self, *args, **kwargs)
        self.parser = None

    def open(self, *args, **kwargs):
        netius.Connection.open(self, *args, **kwargs)
        if not self.is_open(): return
        self.parser = netius.common.SOCKSParser(self)
        self.parser.bind("on_data", self.on_data)
        self.parser.bind("on_auth", self.on_auth)

    def close(self, *args, **kwargs):
        netius.Connection.close(self, *args, **kwargs)
        if not self.is_closed(): return
        if self.parser: self.parser.destroy()

    def send_response(self, status = GRANTED):
        data = struct.pack("!BBHI", 0, status, 0, 0)
        return self.send(data)

    def send_response_extra(self, status = GRANTED_EXTRA):
        version = self.parser.version
        type = self.parser.type
        port = self.parser.port
        address = self.parser.get_address()
        format = "!BBBB%dsH" % len(address)
        data = struct.pack(format, version, status, 0, type, address, port)
        return self.send(data)

    def send_auth(self, version = None, method = 0x00):
        version = version or self.parser.version
        data = struct.pack("!BB", version, method)
        return self.send(data)

    def get_version(self):
        return self.parser.version

    def parse(self, data):
        return self.parser.parse(data)

    def on_data(self):
        self.owner.on_data_socks(self, self.parser)

    def on_auth(self):
        self.owner.on_auth_socks(self, self.parser)

class SOCKSServer(netius.ServerAgent):
    """
    SOCKS base server class to be used as an implementation of the
    RFC 1928 or SOCKSv5 and the SOCKSv4 protocols.

    There are some aspects of the implementation that may not be
    performant driven for readability purposes.
    """

    def __init__(self, rules = {}, throttle = True, max_pending = MAX_PENDING, *args, **kwargs):
        netius.ContainerServer.__init__(
            self,
            receive_buffer_c = int(max_pending * BUFFER_RATIO),
            send_buffer_c = int(max_pending * BUFFER_RATIO),
            *args,
            **kwargs
        ) # @todo how is this going to work (receive buffer control)
        self.rules = rules
        self.throttle = throttle
        self.max_pending = max_pending
        self.min_pending = int(max_pending * MIN_RATIO)
        self.conn_map = {}

        _loop, self.raw_protocol = netius.clients.RawClient.protocol()
        self.raw_protocol.bind("connect", self._on_raw_connect)
        self.raw_protocol.bind("data", self._on_raw_data)
        self.raw_protocol.bind("close", self._on_raw_close)

        #@todo this does not make sense
        self.add_base(self)
        self.add_base(self.raw_client)

    def cleanup(self):
        netius.ContainerServer.cleanup(self)
        self.raw_client.destroy()

    def on_data(self, connection, data):
        netius.ContainerServer.on_data(self, connection, data)

        # tries to retrieve the reference to the tunnel connection
        # currently set in the connection in case it does not exists
        # (initial handshake) runs the parse step on the data and then
        # returns immediately (not going to send it back)
        tunnel_c = hasattr(connection, "tunnel_c") and connection.tunnel_c
        if not tunnel_c: connection.parse(data); return

        # verifies that the current size of the pending buffer is greater
        # than the maximum size for the pending buffer the read operations
        # if that the case the read operations must be disabled
        should_disable = self.throttle and tunnel_c.is_exhausted()
        if should_disable: connection.disable_read()

        # performs the sending operation on the data but uses the throttle
        # callback so that the connection read operations may be resumed if
        # the buffer has reached certain (minimum) levels
        tunnel_c.send(data, callback = self._throttle)

    def on_data_socks(self, connection, parser):
        host = parser.get_host()
        port = parser.port

        _connection = self.raw_client.connect(host, port)
        _connection.max_pending = self.max_pending
        _connection.min_pending = self.min_pending
        connection.tunnel_c = _connection
        self.conn_map[_connection] = connection

    def on_auth_socks(self, connection, parser):
        auth_methods = parser.auth_methods

        if not 0 in auth_methods:
            raise netius.ParserError("Authentication is not supported")

        connection.send_auth(method = 0)

    def on_connection_d(self, connection):
        netius.ContainerServer.on_connection_d(self, connection)

        tunnel_c = hasattr(connection, "tunnel_c") and connection.tunnel_c

        if tunnel_c: tunnel_c.close()

        setattr(connection, "tunnel_c", None)

    def build_connection(self, socket, address, ssl = False):
        return SOCKSConnection(
            owner = self,
            socket = socket,
            address = address,
            ssl = ssl,
            max_pending = self.max_pending,
            min_pending = self.min_pending
        )

    def _throttle(self, _connection):
        if not _connection.is_restored(): return
        connection = self.conn_map[_connection]
        if not connection.renable == False: return
        connection.enable_read()
        self.reads((connection.socket,), state = False)

    def _raw_throttle(self, connection):
        if not connection.is_restored(): return

        tunnel_c = hasattr(connection, "tunnel_c") and connection.tunnel_c
        if not tunnel_c: return
        if not tunnel_c.renable == False: return

        tunnel_c.enable_read()
        self.raw_client.reads((tunnel_c.socket,), state = False)

    def _on_raw_connect(self, client, _connection):
        connection = self.conn_map[_connection]
        version = connection.get_version()
        if version == 0x04: connection.send_response(status = GRANTED)
        elif version == 0x05: connection.send_response_extra(status = GRANTED_EXTRA)

    def _on_raw_data(self, client, _connection, data):
        connection = self.conn_map[_connection]
        should_disable = self.throttle and connection.is_exhausted()
        if should_disable: _connection.disable_read()
        connection.send(data, callback = self._raw_throttle)

    def _on_raw_close(self, client, _connection):
        connection = self.conn_map[_connection]
        connection.close(flush = True)
        del self.conn_map[_connection]

if __name__ == "__main__":
    import logging
    server = SOCKSServer(level = logging.INFO)
    server.serve(env = True)
else:
    __path__ = []



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import netius

class Parser(netius.Observable):

    FIELDS = ("_pid",)

    def __init__(self, owner):
        netius.Observable.__init__(self)
        self.owner = owner
        self._pid = 0

    @classmethod
    def mock(cls, owner, state):
        mock = cls(owner)
        mock.set_state(state)
        return mock

    def build(self):
        netius.Observable.build(self)

    def destroy(self):
        netius.Observable.destroy(self)
        self.owner = None
        self._pid = 0

    def get_state(self):
        cls = self.__class__
        fields = cls.FIELDS
        state = dict()
        for field in fields:
            state[field] = getattr(self, field)
        return state

    def set_state(self, state):
        cls = self.__class__
        fields = cls.FIELDS
        for field in fields:
            value = state[field]
            setattr(self, field, value)

    def info_dict(self):
        info = self.get_state()
        return info

    def parse(self, data):
        self._pid = (self._pid + 1) % 2147483647



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

from . import base

class AddressAuth(base.Auth):

    def __init__(self, allowed = [], *args, **kwargs):
        base.Auth.__init__(self, *args, **kwargs)
        self.allowed = allowed

    @classmethod
    def auth(cls, allowed = [], *args, **kwargs):
        import netius.common
        host = kwargs.get("host", None)
        headers = kwargs.get("headers", {})
        if not host and not headers: return False
        address = headers.get("X-Forwarded-For", host)
        address = headers.get("X-Client-IP", address)
        address = headers.get("X-Real-IP", address)
        address = address.split(",", 1)[0].strip()
        return netius.common.assert_ip4(
            address,
            allowed,
            default = False
        )

    @classmethod
    def is_simple(cls):
        return True

    def auth_i(self, *args, **kwargs):
        return self.__class__.auth(
            allowed = self.allowed,
            *args,
            **kwargs
        )



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import struct

import netius.common

class DHTRequest(netius.Request):

    def __init__(
        self,
        peer_id,
        host = "127.0.0.1",
        port = 9090,
        type = "ping",
        callback = None,
        *args,
        **kwargs
    ):
        netius.Request.__init__(self, callback = callback)
        self.peer_id = peer_id
        self.host = host
        self.port = port
        self.type = type
        self.args = args,
        self.kwargs = kwargs
        self._peer_id = self._get_peer_id()

    @classmethod
    def contact(cls, host, port):
        addr = netius.common.ip4_to_addr(host)
        return struct.pack("!LH", addr, port)

    def request(self):
        if not hasattr(self, self.type):
            raise netius.ParserError("Invalid type '%s'" % self.type)
        method = getattr(self, self.type)
        query = method()
        request = dict(
            t = str(self.id),
            y = "q",
            q = self.type,
            a = query
        )
        return netius.common.bencode(request)

    def ping(self):
        return dict(id = self._peer_id)

    def find_node(self):
        return dict(
            id = self._peer_id,
            target = self.kwargs["target"]
        )

    def get_peers(self):
        return dict(
            id = self._peer_id,
            info_hash = self.kwargs["info_hash"]
        )

    def announce_peer(self):
        return dict(
            id = self._peer_id,
            implied_port = self.kwargs["implied_port"],
            info_hash = self.kwargs["info_hash"],
            port = self.kwargs["port"],
            token = self.kwargs["token"]
        )

    def _get_peer_id(self):
        contact = DHTRequest.contact(self.host, self.port)
        peer_id = netius.legacy.bytes(self.peer_id)
        return peer_id + contact

class DHTResponse(netius.Response):

    def __init__(self, data):
        netius.Response.__init__(self, data)
        self.info = {}

    def parse(self):
        self.info = netius.common.bdecode(self.data)

    def get_id(self):
        t = self.info.get("t", -1)
        return int(t)

    def get_payload(self):
        return self.info.get("r", {})

    def is_error(self):
        return self.info("y", True)

    def is_response(self):
        return self.info("r", True)

class DHTClient(netius.DatagramClient):
    """
    Implementation of the DHT (Distributed hash table) for the torrent
    protocol as the defined in the official specification.

    This implementation is meant to be used in an asynchronous environment
    for maximum performance.

    :see: http://www.bittorrent.org/beps/bep_0005.html
    """

    def ping(self, host, port, peer_id, *args, **kwargs):
        return self.query(type = "ping", *args, **kwargs)

    def find_node(self, *args, **kwargs):
        return self.query(type = "find_node", *args, **kwargs)

    def get_peers(self, *args, **kwargs):
        return self.query(type = "get_peers", *args, **kwargs)

    def query(
        self,
        host = "127.0.0.1",
        port = 9090,
        peer_id = None,
        type = "ping",
        callback = None,
        *args,
        **kwargs
    ):
        request = DHTRequest(
            peer_id,
            host = host,
            port = port,
            type = type,
            callback = callback,
            *args,
            **kwargs
        )
        data = request.request()

        self.add_request(request)

        address = (host, port)
        self.send(data, address)

    def on_data(self, address, data):
        netius.DatagramClient.on_data(self, address, data)

        response = DHTResponse(data)
        response.parse()

        self.on_data_dht(address, response)

    def on_data_dht(self, address, response):
        request = self.get_request(response)
        response.request = request
        self.remove_request(request)

        if not request.callback: return
        request.callback(response)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import re
import time
import base64
import hashlib
import datetime

import netius

from . import asn
from . import rsa
from . import util
from . import mime

def dkim_sign(
    message,
    selector,
    domain,
    private_key,
    identity = None,
    separator = ":",
    creation = None
):
    separator = netius.legacy.bytes(separator)
    identity = identity or "@" + domain

    headers, body = mime.rfc822_parse(message, strip = False)

    if not identity.endswith(domain):
        raise netius.GeneratorError("Identity must end with domain")

    headers = dkim_headers(headers)
    body = dkim_body(body)

    include_headers = [name.lower() for name, _value in headers]
    sign_headers = [header for header in headers if header[0].lower() in include_headers]
    sign_names = [name for name, _value in sign_headers]

    hash = hashlib.sha256()
    hash.update(body)

    body_digest = hash.digest()
    body_hash = base64.b64encode(body_digest)
    body_hash = netius.legacy.str(body_hash)

    creation = creation or time.time()
    creation = int(creation)
    creation_s = str(creation)

    names = separator.join(sign_names)
    names = netius.legacy.str(names)

    sign_fields = [
        ("v", "1"),
        ("a", "rsa-sha256"),
        ("c", "simple/simple"),
        ("d", domain),
        ("i", identity),
        ("l", len(body)),
        ("q", "dns/txt"),
        ("s", selector),
        ("t", creation_s),
        ("h", names),
        ("bh", body_hash),
        ("b", ""),
    ]

    signature = "DKIM-Signature: " + "; ".join("%s=%s" % field for field in sign_fields)
    if isinstance(signature, netius.legacy.UNICODE):
        signature = signature.encode("utf-8")
    signature = dkim_fold(signature)

    hash = hashlib.sha256()
    for name, value in sign_headers:
        hash.update(name)
        hash.update(b":")
        hash.update(value + b"\r\n")

    hash.update(signature)
    digest = hash.digest()

    digest_info = asn.asn1_gen(
        (asn.SEQUENCE, [
            (asn.SEQUENCE, [
                (asn.OBJECT_IDENTIFIER, asn.HASHID_SHA256),
                (asn.NULL, None),
            ]),
            (asn.OCTET_STRING, digest),
        ])
    )

    modulus = private_key["modulus"]
    exponent = private_key["private_exponent"]
    modulus_s = util.integer_to_bytes(modulus)
    modulus_l = len(modulus_s)

    digest_l = len(digest_info)
    delta_l = modulus_l - digest_l - 3
    delta_l = 0 if delta_l < 0 else delta_l

    if digest_l + 3 > modulus_l:
        raise netius.GeneratorError("Hash too large for modulus")

    base = b"\x00\x01" + b"\xff" * delta_l + b"\x00" + digest_info
    base_i = util.bytes_to_integer(base)

    signature_i = rsa.rsa_crypt(base_i, exponent, modulus)
    signature_s = util.integer_to_bytes(signature_i, length = modulus_l)

    signature += base64.b64encode(signature_s) + b"\r\n"
    return signature

def dkim_headers(headers):
    # returns the headers exactly the way they were parsed
    # as this is the simple strategy approach
    return headers

def dkim_body(body):
    # remove the complete set of empty lines in the body
    # and adds only one line to the end of it as requested
    return re.sub(b"(\r\n)*$", b"\r\n", body)

def dkim_fold(header, length = 72):
    """
    Folds a header line into multiple line feed separated lines
    at column length defined (defaults to 72).

    This is required so that the header field is defined according
    to the dkim rules and the default mime encoding.

    :type header: String
    :param header: The string value of the header that is going to
    be folded into multiple lines.
    :type length: int
    :param length: The maximum length of a column until it gets
    broken into multiple lines (in case it's possible).
    :rtype: String
    :return: The folded string value for the header after the correct
    processing of the string value.
    """

    index = header.rfind(b"\r\n ")
    if index == -1: pre = b""
    else:
        index += 3
        pre = header[:index]
        header = header[index:]

    while len(header) > length:
        index = header[:length].rfind(b" ")
        if index == -1: _index = index
        else: _index = index + 1
        pre += header[:index] + b"\r\n "
        header = header[_index:]

    return pre + header

def dkim_generate(domain, suffix = None, number_bits = 1024):
    date_time = datetime.datetime.utcnow()

    selector = date_time.strftime("%Y%m%d%H%M%S")
    if suffix: selector += "." + suffix

    selector_full = "%s._domainkey.%s." % (selector, domain)

    private_key = rsa.rsa_private(number_bits)
    rsa.assert_private(private_key, number_bits = number_bits)
    public_key = rsa.private_to_public(private_key)

    buffer = netius.legacy.BytesIO()
    try:
        rsa.write_private_key(buffer, private_key)
        private_pem = buffer.getvalue()
        private_pem = netius.legacy.str(private_pem)
    finally:
        buffer.close()

    public_data = rsa.asn_public_key(public_key)
    public_b64 = base64.b64encode(public_data)
    public_b64 = netius.legacy.str(public_b64)

    dns_txt = "%s IN TXT \"k=rsa; p=%s\"" % (selector_full, public_b64)

    return dict(
        selector = selector,
        selector_full = selector_full,
        private_pem = private_pem,
        dns_txt = dns_txt
    )



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

from . import request

from .conn import * #@UnusedWildImport
from .common import * #@UnusedWildImport

BUFFER_SIZE = None
""" The size of the buffer that is going to be used in the
sending and receiving of packets from the client, this value
may influence performance by a large factor """

GC_TIMEOUT = 30.0
""" The timeout to be used for the running of the garbage
collector of pending request in a datagram client, this
value will be used n the delay operation of the action """

class Client(Base):
    """
    Abstract client implementation, should provide the required
    mechanisms for basic socket client handling and thread starting
    and managing techniques. Proper and concrete implementation for
    the various socket types should inherit from this class.
    """

    _client = None
    """ The global static client meant to be reused by the
    various static clients that may be created, this client
    may leak creating blocking threads that will prevent the
    system from exiting correctly, in order to prevent that
    the cleanup method should be called """

    def __init__(self, thread = True, daemon = False, *args, **kwargs):
        Base.__init__(self, *args, **kwargs)
        self.receive_buffer = kwargs.get("receive_buffer", BUFFER_SIZE)
        self.send_buffer = kwargs.get("send_buffer", BUFFER_SIZE)
        self.thread = thread
        self.daemon = daemon
        self._thread = None

    @classmethod
    def get_client_s(cls, *args, **kwargs):
        if cls._client: return cls._client
        cls._client = cls(*args, **kwargs)
        return cls._client

    @classmethod
    def cleanup_s(cls):
        if not cls._client: return
        cls._client.close()

    def ensure_loop(self, env = True):
        """
        Ensures that the proper main loop thread requested in the building
        of the entity is started if that was requested.

        This mechanism is required because the thread construction and
        starting should be deferred until an operation in the connection
        is requested (lazy thread construction).

        The call to this method should be properly inserted on the code so
        that it is only called when a main (polling) loop is required.

        :type env: bool
        :param env: If the environment variables should be used for the
        setting of some of the parameters of the new client/poll to be used.
        """

        # verifies if the (run in) thread flag is set and that the there's
        # not thread currently created for the client in case any of these
        # conditions fails the control flow is returned immediately to caller
        if not self.thread: return
        if self._thread: return

        # runs the various extra variable initialization taking into
        # account if the environment variable is currently set or not
        # please note that some side effects may arise from this set
        if env: self.level = self.get_env("LEVEL", self.level)
        if env: self.diag = self.get_env("CLIENT_DIAG", self.diag, cast = bool)
        if env: self.logging = self.get_env("LOGGING", self.logging)
        if env: self.poll_name = self.get_env("POLL", self.poll_name)
        if env: self.poll_timeout = self.get_env(
            "POLL_TIMEOUT",
            self.poll_timeout,
            cast = float
        )

        # prints a debug message about the new thread to be created for
        # the client infra-structure (required for execution)
        self.debug("Starting new thread for '%s' ..." % self.name)

        # in case the thread flag is set a new thread must be constructed
        # for the running of the client's main loop then, these thread
        # may or may not be constructed using a daemon approach
        self._thread = BaseThread(
            owner = self,
            daemon = self.daemon,
            name = "Loop"
        )
        self._thread.start()

    def join(self, timeout = None):
        # runs the join operation in the thread associated with the client
        # so that the current thread blocks until the other ends execution
        self._thread.join(timeout = timeout)

class DatagramClient(Client):

    def __init__(self, *args, **kwargs):
        Client.__init__(self, *args, **kwargs)
        self.socket = None
        self.renable = True
        self.wready = True
        self.pending_s = 0
        self.pending = collections.deque()
        self.requests = []
        self.requests_m = {}
        self.pending_lock = threading.RLock()

    def boot(self):
        Client.boot(self)
        self.keep_gc(timeout = GC_TIMEOUT, run = False)

    def cleanup(self):
        Client.cleanup(self)
        del self.requests[:]
        self.requests_m.clear()

    def on_read(self, _socket):
        callbacks = self.callbacks_m.get(_socket, None)
        if callbacks:
            for callback in callbacks: callback("read", _socket)

        # verifies if the provided socket for reading is the same
        # as the one registered in the client if that's not the case
        # return immediately to avoid unwanted operations
        if not _socket == self.socket: return

        try:
            # iterates continuously trying to read as much data as possible
            # when there's a failure to read more data it should raise an
            # exception that should be handled properly
            while True:
                data, address = _socket.recvfrom(CHUNK_SIZE)
                self.on_data(address, data)
        except ssl.SSLError as error:
            error_v = error.args[0] if error.args else None
            error_m = error.reason if hasattr(error, "reason") else None
            if error_v in SSL_SILENT_ERRORS:
                self.on_expected(error)
            elif not error_v in SSL_VALID_ERRORS and\
                not error_m in SSL_VALID_REASONS:
                self.on_exception(error)
        except socket.error as error:
            error_v = error.args[0] if error.args else None
            if error_v in SILENT_ERRORS:
                self.on_expected(error)
            elif not error_v in VALID_ERRORS:
                self.on_exception(error)
        except (KeyboardInterrupt, SystemExit):
            raise
        except BaseException as exception:
            self.on_exception(exception)

    def on_write(self, _socket):
        callbacks = self.callbacks_m.get(_socket, None)
        if callbacks:
            for callback in callbacks: callback("write", _socket)

        # verifies if the provided socket for writing is the same
        # as the one registered in the client if that's not the case
        # return immediately to avoid unwanted operations
        if not _socket == self.socket: return

        try:
            self._send(_socket)
        except ssl.SSLError as error:
            error_v = error.args[0] if error.args else None
            error_m = error.reason if hasattr(error, "reason") else None
            if error_v in SSL_SILENT_ERRORS:
                self.on_expected(error)
            elif not error_v in SSL_VALID_ERRORS and\
                not error_m in SSL_VALID_REASONS:
                self.on_exception(error)
        except socket.error as error:
            error_v = error.args[0] if error.args else None
            if error_v in SILENT_ERRORS:
                self.on_expected(error)
            elif not error_v in VALID_ERRORS:
                self.on_exception(error)
        except (KeyboardInterrupt, SystemExit):
            raise
        except BaseException as exception:
            self.on_exception(exception)

    def on_error(self, _socket):
        callbacks = self.callbacks_m.get(_socket, None)
        if callbacks:
            for callback in callbacks: callback("error", _socket)

        # verifies if the provided socket for error is the same
        # as the one registered in the client if that's not the case
        # return immediately to avoid unwanted operations
        if not _socket == self.socket: return

    def on_exception(self, exception):
        self.warning(exception)
        self.log_stack()

    def on_expected(self, exception):
        self.debug(exception)

    def on_data(self, connection, data):
        pass

    def keep_gc(self, timeout = GC_TIMEOUT, run = True):
        if run: self.gc()
        self.delay(self.keep_gc, timeout)

    def gc(self, callbacks = True):
        # in case there're no requests pending in the current client
        # there's no need to start the garbage collection logic, as
        # this would required some (minimal) resources
        if not self.requests: return

        # prints a message (for debug) about the garbage collection
        # operation that is going to be run
        self.debug("Running garbage collection ...")

        # retrieves the current time value and iterates over the
        # various request currently defined in the client (pending
        # and answer) to try to find the ones that have time out
        current = time.time()
        while True:
            # verifies if the requests structure (list) is empty and
            # if that's the case break the loop, nothing more remains
            # to be processed for the current garbage collection operation
            if not self.requests: break

            # retrieves the top level request (peek operation) and
            # verifies if the timeout value of it has exceed the
            # current time if that's the case removes it as it
            # should no longer be handled (time out)
            request = self.requests[0]
            if request.timeout > current: break
            self.remove_request(request)

            # in case the (call) callbacks flag is not set continues
            # the current loop, meaning that the associated callbacks
            # are not going to be called (as expected)
            if not callbacks: continue

            # extracts the callback method from the request and in
            # case it is defined and valid calls it with an invalid
            # argument meaning that an error has occurred, note that
            # the call is encapsulated in a safe manner meaning that
            # any exception raised will be gracefully caught
            callback = request.callback
            callback and self.call_safe(callback, args = [None])

    def add_request(self, request):
        # adds the current request object to the list of requests
        # that are pending a valid response, a garbage collector
        # system should be able to erase this request from the
        # pending list in case a timeout value has passed
        self.requests.append(request)
        self.requests_m[request.id] = request

    def remove_request(self, request):
        self.requests.remove(request)
        del self.requests_m[request.id]

    def get_request(self, id):
        is_response = isinstance(id, request.Response)
        if is_response: id = id.get_id()
        return self.requests_m.get(id, None)

    def ensure_socket(self):
        # in case the socket is already created and valid returns immediately
        # as nothing else remain to be done in the current method
        if self.socket: return

        # prints a small debug message about the udp socket that is going
        # to be created for the client's connection
        self.debug("Creating clients's udp socket ...")

        # creates the socket that it's going to be used for the listening
        # of new connections (client socket) and sets it as non blocking
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.socket.setblocking(0)

        # sets the various options in the service socket so that it becomes
        # ready for the operation with the highest possible performance
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)

        # adds the socket to all of the pool lists so that it's ready to read
        # write and handle error, this is the expected behavior of a service
        # socket so that it can handle all of the expected operations
        self.sub_all(self.socket)

    def ensure_write(self):
        # retrieves the identifier of the current thread and
        # checks if it's the same as the one defined in the
        # owner in case it's not then the operation is not
        # considered to be safe and must be delayed
        cthread = threading.current_thread()
        tid = cthread.ident or 0
        is_safe = tid == self.tid

        # in case the thread where this code is being executed
        # is not the same the operation is considered to be not
        # safe and so it must be delayed to be executed in the
        # next loop of the thread cycle, must return immediately
        # to avoid extra subscription operations
        if not is_safe: return self.delay(self.ensure_write, safe = True)

        # adds the current socket to the list of write operations
        # so that it's going to be available for writing as soon
        # as possible from the poll mechanism
        self.sub_write(self.socket)

    def remove_write(self):
        self.unsub_write(self.socket)

    def enable_read(self):
        if not self.renable == False: return
        self.renable = True
        self.sub_read(self.socket)

    def disable_read(self):
        if not self.renable == True: return
        self.renable = False
        self.unsub_read(self.socket)

    def send(
        self,
        data,
        address,
        delay = True,
        ensure_loop = True,
        callback = None
    ):
        if ensure_loop: self.ensure_loop()

        data = legacy.bytes(data)
        data_l = len(data)

        if callback: data = (data, callback)
        data = (data, address)

        cthread = threading.current_thread()
        tid = cthread.ident or 0
        is_safe = tid == self.tid

        self.pending_lock.acquire()
        try: self.pending.appendleft(data)
        finally: self.pending_lock.release()

        self.pending_s += data_l

        if self.wready:
            if is_safe and not delay: self._flush_write()
            else: self.delay(
                self._flush_write,
                immediately = True,
                verify = True,
                safe = True
            )
        else:
            self.ensure_write()

    def _send(self, _socket):
        self.wready = True
        self.pending_lock.acquire()
        try:
            while True:
                # in case there's no pending data to be sent to the
                # server side breaks the current loop (queue empty)
                if not self.pending: break

                # retrieves the current data from the pending list
                # of data to be sent and then saves the original data
                # object (for latter usage), sets the callback as not
                # defined and then unpacks the data into data and address
                data = self.pending.pop()
                data_o = data
                callback = None
                data, address = data

                # verifies if the data type of the data is a tuple and
                # if that's the case unpacks it as data and callback
                is_tuple = type(data) == tuple
                if is_tuple: data, callback = data

                # retrieves the length (in bytes) of the data that is
                # going to be sent to the server
                data_l = len(data)

                try:
                    # tries to send the data through the socket and
                    # retrieves the number of bytes that were correctly
                    # sent through the socket, this number may not be
                    # the same as the size of the data in case only
                    # part of the data has been sent
                    if data: count = _socket.sendto(data, address)
                    else: count = 0

                    # verifies if the current situation is that of a non
                    # closed socket and valid data, and if that's the case
                    # and no data has been sent the socket is considered to
                    # be in a would block situation and and such an error
                    # is raised indicating the issue (is going to be caught
                    # as a normal would block exception)
                    if data and count == 0: raise socket.error(errno.EWOULDBLOCK)
                except:
                    # sets the current connection write ready flag to false
                    # so that a new level notification must be received
                    self.wready = False

                    # ensures that the write event is going to be triggered
                    # this is required so that the remaining pending data is
                    # going to be correctly written on a new write event,
                    # triggered when the connection is ready for more writing
                    self.ensure_write()

                    # in case there's an exception must add the data
                    # object to the list of pending data because the data
                    # has not been correctly sent
                    self.pending.append(data_o)
                    raise
                else:
                    # decrements the size of the pending buffer by the number
                    # of bytes that were correctly send through the buffer
                    self.pending_s -= count

                    # verifies if the data has been correctly sent through
                    # the socket and for suck situations calls the callback
                    # object, otherwise creates a new data object with only
                    # the remaining (partial data) and the callback to be
                    # sent latter (only then the callback is called)
                    is_valid = count == data_l
                    if is_valid:
                        callback and callback(self)
                    else:
                        data_o = ((data[count:], callback), address)
                        self.pending.append(data_o)
        finally:
            self.pending_lock.release()

        self.remove_write()

    def _flush_write(self):
        """
        Flush operations to be called by the delaying controller
        (in ticks) that will trigger all the write operations
        pending for the current connection's socket.
        """

        self.ensure_socket()
        self.writes((self.socket,), state = False)

class StreamClient(Client):

    def __init__(self, *args, **kwargs):
        Client.__init__(self, *args, **kwargs)
        self.pendings = []
        self.free_map = {}
        self._pending_lock = threading.RLock()

    def cleanup(self):
        Client.cleanup(self)
        del self.pendings[:]
        self.free_map.clear()
        self._pending_lock = None

    def ticks(self):
        self.set_state(STATE_TICK)
        self._lid = (self._lid + 1) % 2147483647
        if self.pendings: self._connects()
        self._delays()

    def info_dict(self, full = False):
        info = Client.info_dict(self, full = full)
        if full: info.update(
            pendings = len(self.pendings),
            free_conn = sum([len(value) for value in legacy.values(self.free_map)])
        )
        return info

    def acquire_c(
        self,
        host,
        port,
        ssl = False,
        key_file = None,
        cer_file = None,
        validate = False,
        callback = None
    ):
        # sets the initial value of the connection instance variable
        # to invalid, this is going to be populated with a valid
        # connection from either the pool of connections or a new one
        connection = None

        # creates the tuple that is going to describe the connection
        # and tries to retrieve a valid connection from the map of
        # free connections (connection re-usage)
        connection_t = (host, port, ssl, key_file, cer_file)
        connection_l = self.free_map.get(connection_t, None)

        # iterates continuously trying to retrieve a valid and open
        # connection from the list of connection that compose the
        # pool of connections for the current client
        while connection_l:
            # retrieves the first connection in the list (pop) and
            # then validates it trying to determine if the connection
            # is still valid (open and ready), if that's not the case
            # unsets the connection variable
            connection = connection_l.pop()
            if validate and not self.validate_c(connection): connection = None

            # in case the connection has been invalidated (possible
            # disconnect) the current loop iteration is skipped and
            # a new connection from the list of connections in pool
            # is going to be searched and validated
            if not connection: continue

            # runs the connection acquire operation that should take
            # care of the proper acquisition notification process and
            # then breaks the cycle (valid connection found)
            self.acquire(connection)
            break

        # in case no connection is found a new one must be created by
        # establishing a connect operation, this operation is not going
        # to be performed immediately as it's going to be deferred to
        # the next execution cycle (delayed execution)
        if not connection:
            connection = self.connect(
                host,
                port,
                ssl = ssl,
                key_file = key_file,
                cer_file = cer_file
            )
            connection.tuple = connection_t

        # returns the connection object the caller method, this connection
        # is acquired and should be safe and ready to be used
        return connection

    def release_c(self, connection):
        if not hasattr(connection, "tuple"): return
        connection_t = connection.tuple
        connection_l = self.free_map.get(connection_t, [])
        connection_l.append(connection)
        self.free_map[connection_t] = connection_l
        self.on_release(connection)

    def remove_c(self, connection):
        if not hasattr(connection, "tuple"): return
        connection_t = connection.tuple
        connection_l = self.free_map.get(connection_t, [])
        if connection in connection_l: connection_l.remove(connection)

    def validate_c(self, connection, close = True):
        # sets the original valid flag value as true so that the
        # basic/default assumption on the connection is that it's
        # valid (per basis a connection is valid)
        valid = True

        # tries to retrieve the value of the error options value of
        # the socket in case it's currently set unsets the valid flag
        error = connection.socket.getsockopt(socket.SOL_SOCKET, socket.SO_ERROR)
        if error: valid = False

        # iterates continuously trying to read any pending data from
        # the connection, some of this data may indicate that the
        # connection is no longer valid for usage
        while True:
            # verifies if the value of the valid flag is false and
            # if that's the case breaks the current loop immediately
            if not valid: break

            # tries to read/receive any set of pending data from
            # the connection in case there's an exception and it's
            # considered valid ignores the exceptions and considers
            # the connection valid (breaks loop) otherwise closes the
            # connection and sets it as invalid
            try:
                data = connection.recv()
                if not data: raise errors.NetiusError("EOF received")
                connection.send(b"")
            except ssl.SSLError as error:
                error_v = error.args[0] if error.args else None
                if error_v in SSL_VALID_ERRORS: break
                if close: connection.close()
                valid = False
            except socket.error as error:
                error_v = error.args[0] if error.args else None
                if error_v in VALID_ERRORS: break
                if close: connection.close()
                valid = False
            except (KeyboardInterrupt, SystemExit):
                raise
            except:
                if close: connection.close()
                valid = False

        # returns the final value on the connection validity test
        # indicating if the connection is ready for usage or not
        return valid

    def connect(
        self,
        host,
        port,
        ssl = False,
        key_file = None,
        cer_file = None,
        ca_file = None,
        ca_root = True,
        ssl_verify = False,
        family = socket.AF_INET,
        type = socket.SOCK_STREAM,
        ensure_loop = True,
        env = True
    ):
        # runs a series of pre-validations on the provided parameters, raising
        # exceptions in case they do not comply with expected values
        if not host: raise errors.NetiusError("Invalid host for connect operation")
        if not port: raise errors.NetiusError("Invalid port for connect operation")

        # tries to retrieve some of the environment variable related values
        # so that some of these values are accessible via an external environment
        # allowing extra configuration flexibility for the client
        key_file = self.get_env("KEY_FILE", key_file) if env else key_file
        cer_file = self.get_env("CER_FILE", cer_file) if env else cer_file
        ca_file = self.get_env("CA_FILE", ca_file) if env else ca_file
        ca_root = self.get_env("CA_ROOT", ca_root, cast = bool) if env else ca_root
        ssl_verify = self.get_env("SSL_VERIFY", ssl_verify, cast = bool) if env else ssl_verify
        key_file = self.get_env("KEY_DATA", key_file, expand = True) if env else key_file
        cer_file = self.get_env("CER_DATA", cer_file, expand = True) if env else cer_file
        ca_file = self.get_env("CA_DATA", ca_file, expand = True) if env else ca_file

        # ensures that a proper loop cycle is available for the current
        # client, otherwise the connection operation would become stalled
        # because there's no listening of events for it
        if ensure_loop: self.ensure_loop()

        # ensures that the proper socket family is defined in case the
        # requested host value is unix socket oriented, this step greatly
        # simplifies the process of created unix socket based clients
        family = socket.AF_UNIX if host == "unix" else family

        # verifies the kind of socket that is going to be used for the
        # connect operation that is going to be performed, note that the
        # unix type should be used with case as it does not exist in every
        # operative system and may raised an undefined exceptions
        is_unix = hasattr(socket, "AF_UNIX") and family == socket.AF_UNIX
        is_inet = family in (socket.AF_INET, socket.AF_INET6)

        # runs a series of default operation for the SSL related attributes
        # that are going to be used in the socket creation and wrapping
        key_file = key_file or SSL_KEY_PATH
        cer_file = cer_file or SSL_CER_PATH
        ca_file = ca_file or SSL_CA_PATH

        # determines if the SSL verify flag value is valid taking into account
        # the provided value and defaulting to false value if not valid
        ssl_verify = ssl_verify or False

        # creates the client socket value using the provided family and socket
        # type values and then sets it immediately as non blocking
        _socket = socket.socket(family, type)
        _socket.setblocking(0)

        # in case the SSL flag is set re-creates the socket by wrapping it into
        # an SSL based one with the provided set of keys and certificates
        if ssl: _socket = self._ssl_wrap(
            _socket,
            key_file = key_file,
            cer_file = cer_file,
            ca_file = ca_file,
            ca_root = ca_root,
            server = False,
            ssl_verify = ssl_verify,
            server_hostname = host
        )

        # sets the appropriate socket options enable it for port re-usage and
        # for proper keep alive notification, among others
        _socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        _socket.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
        if is_inet: _socket.setsockopt(
            socket.IPPROTO_TCP,
            socket.TCP_NODELAY,
            1
        )
        if self.receive_buffer: _socket.setsockopt(
            socket.SOL_SOCKET,
            socket.SO_RCVBUF,
            self.receive_buffer
        )
        if self.send_buffer: _socket.setsockopt(
            socket.SOL_SOCKET,
            socket.SO_SNDBUF,
            self.send_buffer
        )
        self._socket_keepalive(_socket)

        # constructs the address tuple taking into account if the
        # socket is unix based or if instead it represents a "normal"
        # one and the host and port must be used instead
        address = port if is_unix else (host, port)

        # creates the connection object using the typical constructor
        # and then sets the SSL host (for verification) if the verify
        # SSL option is defined (secured and verified connection)
        connection = self.build_connection(_socket, address, ssl = ssl)
        if ssl_verify: connection.ssl_host = host

        # acquires the pending lock so that it's safe to add an element
        # to the list of pending connection for connect, this lock is
        # then released in the final part of the operation
        self._pending_lock.acquire()
        try: self.pendings.append(connection)
        finally: self._pending_lock.release()

        # returns the "final" connection, that is now scheduled for connect
        # to the caller method, it may now be used for operations
        return connection

    def acquire(self, connection):
        acquire = lambda: self.on_acquire(connection)
        self.delay(acquire)

    def on_read(self, _socket):
        # tries to retrieve a possible callback registered for the socket
        # and if there's one calls it to be able to "append" extra operations
        # to the execution of the read operation in the socket
        callbacks = self.callbacks_m.get(_socket, None)
        if callbacks:
            for callback in callbacks: callback("read", _socket)

        # retrieves the connection object associated with the
        # current socket that is going to be read in case there's
        # no connection available or the status is not open
        # must return the control flow immediately to the caller
        connection = self.connections_m.get(_socket, None)
        if not connection: return
        if not connection.status == OPEN: return
        if not connection.renable == True: return

        try:
            # in case the connection is under the connecting state
            # the socket must be verified for errors and in case
            # there's none the connection must proceed, for example
            # the SSL connection handshake must be performed/retried
            if connection.connecting: self._connectf(connection)

            # verifies if there's any pending operations in the
            # connection (eg: SSL handshaking) and performs it trying
            # to finish them, if they are still pending at the current
            # state returns immediately (waits for next loop)
            if self._pending(connection): return

            # iterates continuously trying to read as much data as possible
            # when there's a failure to read more data it should raise an
            # exception that should be handled properly
            while True:
                data = connection.recv(CHUNK_SIZE)
                if data: self.on_data(connection, data)
                else: connection.close(); break
                if not connection.status == OPEN: break
                if not connection.renable == True: break
                if not connection.socket == _socket: break
        except ssl.SSLError as error:
            error_v = error.args[0] if error.args else None
            error_m = error.reason if hasattr(error, "reason") else None
            if error_v in SSL_SILENT_ERRORS:
                self.on_expected(error, connection)
            elif not error_v in SSL_VALID_ERRORS and\
                not error_m in SSL_VALID_REASONS:
                self.on_exception(error, connection)
        except socket.error as error:
            error_v = error.args[0] if error.args else None
            if error_v in SILENT_ERRORS:
                self.on_expected(error, connection)
            elif not error_v in VALID_ERRORS:
                self.on_exception(error, connection)
        except (KeyboardInterrupt, SystemExit):
            raise
        except BaseException as exception:
            self.on_exception(exception, connection)

    def on_write(self, _socket):
        # tries to retrieve a possible callback registered for the socket
        # and if there's one calls it to be able to "append" extra operations
        # to the execution of the read operation in the socket
        callbacks = self.callbacks_m.get(_socket, None)
        if callbacks:
            for callback in callbacks: callback("write", _socket)

        # retrieves the connection associated with the socket that
        # is ready for the write operation and verifies that it
        # exists and the current status of it is open (required)
        connection = self.connections_m.get(_socket, None)
        if not connection: return
        if not connection.status == OPEN: return

        # in case the connection is under the connecting state
        # the socket must be verified for errors and in case
        # there's none the connection must proceed, for example
        # the SSL connection handshake must be performed/retried
        if connection.connecting: self._connectf(connection)

        try:
            connection._send()
        except ssl.SSLError as error:
            error_v = error.args[0] if error.args else None
            error_m = error.reason if hasattr(error, "reason") else None
            if error_v in SSL_SILENT_ERRORS:
                self.on_expected(error, connection)
            elif not error_v in SSL_VALID_ERRORS and\
                not error_m in SSL_VALID_REASONS:
                self.on_exception(error, connection)
        except socket.error as error:
            error_v = error.args[0] if error.args else None
            if error_v in SILENT_ERRORS:
                self.on_expected(error, connection)
            elif not error_v in VALID_ERRORS:
                self.on_exception(error, connection)
        except (KeyboardInterrupt, SystemExit):
            raise
        except BaseException as exception:
            self.on_exception(exception, connection)

    def on_error(self, _socket):
        callbacks = self.callbacks_m.get(_socket, None)
        if callbacks:
            for callback in callbacks: callback("error", _socket)

        connection = self.connections_m.get(_socket, None)
        if not connection: return
        if not connection.status == OPEN: return

        connection.close()

    def on_exception(self, exception, connection):
        self.warning(exception)
        self.log_stack()
        connection.close()

    def on_expected(self, exception, connection):
        self.debug(exception)
        connection.close()

    def on_connect(self, connection):
        self.debug(
            "Connection '%s' from '%s' connected" %
            (connection.id, connection.owner.name)
        )
        connection.set_connected()
        if hasattr(connection, "tuple"):
            self.on_acquire(connection)

    def on_upgrade(self, connection):
        self.debug(
            "Connection '%s' from '%s' upgraded" %
            (connection.id, connection.owner.name)
        )
        connection.set_upgraded()

    def on_ssl(self, connection):
        # runs the connection host verification process for the SSL
        # meaning that in case an SSL host value is defined it is going
        # to be verified against the value in the certificate
        connection.ssl_verify_host()

        # runs the connection fingerprint verification the will try to
        # match the digest of the peer certificate against the one that
        # is expected from it (similar to host verification)
        connection.ssl_verify_fingerprint()

        # verifies if the connection is either connecting or upgrading
        # and calls the proper event handler for each event, this is
        # required because the connection workflow is probably dependent
        # on the calling of these event handlers to proceed
        if connection.connecting: self.on_connect(connection)
        elif connection.upgrading: self.on_upgrade(connection)

    def on_acquire(self, connection):
        pass

    def on_release(self, connection):
        pass

    def on_data(self, connection, data):
        connection.set_data(data)

    def _connectf(self, connection):
        """
        Finishes the process of connecting to the remote end-point
        this should be done in certain steps of the connection.

        The process of finishing the connecting process should include
        the SSL handshaking process.

        :type connection: Connection
        :param connection: The connection that should have the connect
        process tested for finishing.
        """

        # in case the SSL connection is still undergoing the handshaking
        # procedures (marked as connecting) ignores the call as this must
        # be a duplicated call to this method (to be ignored)
        if connection.ssl_connecting: return

        # verifies if there was an error in the middle of the connection
        # operation and if that's the case calls the proper callback and
        # returns the control flow to the caller method
        error = connection.socket.getsockopt(socket.SOL_SOCKET, socket.SO_ERROR)
        if error: self.on_error(connection.socket); return

        # checks if the current connection is SSL based and if that's the
        # case starts the handshaking process (async non blocking) otherwise
        # calls the on connect callback with the newly created connection
        if connection.ssl: connection.add_starter(self._ssl_handshake)
        else: self.on_connect(connection)

        # runs the starter process (initial kick-off) so that all the starters
        # registered for the connection may start to be executed, note that if
        # the SSL handshake starter has been registered its first execution is
        # going to be triggered by this call
        connection.run_starter()

    def _connects(self):
        self._pending_lock.acquire()
        try:
            while self.pendings:
                connection = self.pendings.pop()
                self._connect(connection)
        finally:
            self._pending_lock.release()

    def _connect(self, connection):
        # in case the current connection has been closed meanwhile
        # the current connection is meant to be avoided and so the
        # method must return immediately to the caller method
        if connection.status == CLOSED: return

        # retrieves the socket associated with the connection
        # and calls the open method of the connection to proceed
        # with the correct operations for the connection
        _socket = connection.socket
        connection.open(connect = True)

        # tries to run the non blocking connection it should
        # fail and the connection should only be considered as
        # open when a write event is raised for the connection
        try: _socket.connect(connection.address)
        except ssl.SSLError as error:
            error_v = error.args[0] if error.args else None
            if not error_v in SSL_VALID_ERRORS:
                self.warning(error)
                self.log_stack()
                self.trigger("error", self, connection, error)
                connection.close()
                return
        except socket.error as error:
            error_v = error.args[0] if error.args else None
            if not error_v in VALID_ERRORS:
                self.warning(error)
                self.log_stack()
                self.trigger("error", self, connection, error)
                connection.close()
                return
        except (KeyboardInterrupt, SystemExit):
            raise
        except BaseException as exception:
            self.warning(exception)
            self.log_stack()
            self.trigger("error", self, connection, exception)
            connection.close()
            raise

        # otherwise the connect operation has finished correctly
        # and the finish connect method should be called indicating
        # that the connect operation has completed successfully
        else:
            self._connectf(connection)

        # in case the connection is not of type SSL the method
        # may return as there's nothing left to be done, as the
        # rest of the method is dedicated to SSL tricks
        if not connection.ssl: return

        # verifies if the current SSL object is a context oriented one
        # (newest versions) or a legacy oriented one, that does not uses
        # any kind of context object, this is relevant in order to make
        # decisions on how the SSL object may be re-constructed
        has_context = hasattr(_socket, "context")
        has_sock = hasattr(_socket, "_sock")

        # creates the SSL object for the socket as it may have been
        # destroyed by the underlying SSL library (as an error) because
        # the socket is of type non blocking and raises an error, note
        # that the creation of the socket varies between SSL versions
        if _socket._sslobj: return
        if has_context:
            _socket._sslobj = _socket.context._wrap_socket(
                _socket,
                _socket.server_side,
                _socket.server_hostname
            )
        else:
            _socket._sslobj = ssl._ssl.sslwrap(
                _socket._sock if has_sock else _socket,
                False,
                _socket.keyfile,
                _socket.certfile,
                _socket.cert_reqs,
                _socket.ssl_version,
                _socket.ca_certs
            )

        # verifies if the SSL object class is defined in the SSL module
        # and if that's the case an extra wrapping operation is performed
        # in order to comply with new indirection/abstraction method, under
        # some circumstances this operations fails with an exception because
        # the wrapping operation is not allowed for every Python environment
        if not hasattr(ssl, "SSLObject"): return
        try: _socket._sslobj = ssl.SSLObject(_socket._sslobj, owner = _socket)
        except TypeError: pass

    def _ssl_handshake(self, connection):
        Client._ssl_handshake(self, connection)

        # verifies if the socket still has finished the SSL handshaking
        # process (by verifying the appropriate flag) and then if that's
        # not the case returns immediately (nothing done)
        if not connection.ssl_handshake: return

        # prints a debug information notifying the developer about
        # the finishing of the handshaking process for the connection
        self.debug("SSL Handshaking completed for connection")

        # calls the proper callback on the connection meaning
        # that SSL is now enabled for that socket/connection and so
        # the communication between peers is now secured
        self.on_ssl(connection)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import inspect

import logging.handlers

SILENT = logging.CRITICAL + 1
""" The "artificial" silent level used to silent a logger
or an handler, this is used as an utility for debugging
purposes more that a real feature for production systems """

def rotating_handler(
    path = "netius.log",
    max_bytes = 1048576,
    max_log = 5,
    encoding = None,
    delay = False
):
    return logging.handlers.RotatingFileHandler(
        path,
        maxBytes = max_bytes,
        backupCount = max_log,
        encoding = encoding,
        delay = delay
    )

def smtp_handler(
    host = "localhost",
    port = 25,
    sender = "no-reply@netius.com",
    receivers = [],
    subject = "Netius logging",
    username = None,
    password = None,
    stls = False
):
    address = (host, port)
    if username and password: credentials = (username, password)
    else: credentials = None
    has_secure = in_signature(logging.handlers.SMTPHandler.__init__, "secure")
    if has_secure: kwargs = dict(secure = () if stls else None)
    else: kwargs = dict()
    return logging.handlers.SMTPHandler(
        address,
        sender,
        receivers,
        subject,
        credentials = credentials,
        **kwargs
    )

def in_signature(callable, name):
    has_full = hasattr(inspect, "getfullargspec")
    if has_full: spec = inspect.getfullargspec(callable)
    else: spec = inspect.getargspec(callable)
    args, _varargs, kwargs = spec[:3]
    return (args and name in args) or (kwargs and "secure" in kwargs)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import math
import random

import netius

from . import util

def prime(number_bits):
    """
    Generates a prime number with the given number of bits
    in length. This is a brute force based generation as a
    random number is generated and then tested for primality.

    :type number_bits: int
    :param number_bits: The number of bits to be used in
    the prime number generation.
    :rtype: int
    :return: The generated prime number, that should have
    been verified for primality.
    """

    # iterates continuously, trying to find a large enough
    # prime number as requested by the call
    while True:
        # generates a random number and then makes sure that
        # it's an odd number (last bit to one)
        integer = util.random_integer(number_bits)
        integer |= 1

        # verifies if the generated integer value is a prime
        # using the primality testing strategy, and in case
        # it's breaks the current loop as a prime has been
        # found with the pre-defined number of bits
        if is_prime(integer): break

    # returns the (generated) and verified prime integer
    # to the caller method, may be used for exponent
    return integer

def is_prime(number):
    return random_primality(number, 6)

def relatively_prime(first, second):
    # retrieves the greatest common divisor between the
    # two values and verifies if the value is one, for
    # such situations they are "relative primes"
    divisor = gcd(first, second)
    return divisor == 1

def gcd(first, second):
    """
    Calculates the greatest common divisor of p value and q value.
    This method uses the classic euclidean algorithm.

    :type first: int
    :param first: The first number to obtain the greatest
    common divisor using the euclidean algorithm.
    :type second: int
    :param second: The second number to obtain the greatest
    common divisor using the euclidean algorithm.
    :rtype: int
    :return: The greatest common divisor between both values.
    :see: http://en.wikipedia.org/wiki/Euclidean_algorithm
    """

    # in case the p value is smaller than the q value
    # reverses the order of the arguments and re-computes
    if first < second: return gcd(second, first)

    # in case the q value is zero
    if second == 0:
        # returns the p value
        # because there is no division by zero
        return first

    # calculates the next "second" value that is going
    # to be used in next iteration and runs the next
    # iteration for those calculus
    next = abs(first % second)
    return gcd(second, next)

def egcd(first, second):
    """
    Extended version of the greatest common divisor created
    by Euclid. Computes additional coefficients of the
    Bézout's identity.

    :type first: int
    :param first: The first number to obtain the greatest
    common divisor using the euclidean algorithm.
    :type second: int
    :param second: The second number to obtain the greatest
    common divisor using the euclidean algorithm.
    :rtype: Tuple
    :return: A tuple containing the various coefficients calculated
    for this extended approach of the greatest common divisor.
    :see: http://en.wikipedia.org/wiki/Extended_Euclidean_algorithm
    """

    if second == 0: return (first, 1, 0)

    q = abs(first % second)
    r = first // second
    d, k, l = egcd(second, q)

    return (d, l, k - l * r)

def modinv(first, second):
    """
    Uses the extended greatest common divisor algorithm to compute
    the modulus of an inverted value against another.

    The execution of this method is equivalent to (1 / first mod second)
    using mathematical terms.

    :type first: int
    :param first: The first value, that is going to be inverted before
    the modulus operation is performed.
    :type second: int
    :param second: The second value that is going to be used as the basis
    for the modulus operation.
    :rtype: int
    :return: The result of the computation of inverted modulus according
    to its mathematical definition.
    """

    d, l, _e = egcd(first, second)
    if d != 1: raise netius.DataError("Modular inverse does not exist")
    else: return l % second

def random_integer_interval(min_value, max_value):
    # sets the default minimum number of bits, even if the
    # range is too small, (represents integer value)
    max_number_bits = 32

    # calculates the range of the random numbers to generate,
    # meaning the amount of numbers that may be generated for
    # the currently defined domain and then converts this value
    # into both bits and bytes (different math radix)
    range = max_value - min_value
    range_bits = math.log(range, 2)
    range_bytes = ceil_integer(range_bits / 8.0)

    # converts the range into bits, but verifies that there
    # is at least a minimum number of bits
    range_bits = max(range_bytes * 8, max_number_bits * 2)

    # generates the random number of bits to be used in the
    # number generation, taking into account that the value
    # of it will never be greater that the maximum values
    number_bits = random.randint(max_number_bits, range_bits)

    # generates the random integer with the number of bits generated
    # and applying modulo of the range, then increments the minimum
    # value to the generated value and returns it to the caller
    random_base_value = util.random_integer(number_bits) % range
    return random_base_value + min_value

def random_primality(number, k):
    """
    Uses a probabilistic approach to the testing of primality
    of a number. The resulting value has an error probability
    of (2 ** -k), so a k value should be chosen taking a low
    error as target.

    :type number: int
    :param number: The value that is going to be tested for
    primality.
    :type k: int
    :param k: The coefficient that is going to be used in the
    test, the higher this value is the small the error is.
    :see: http://en.wikipedia.org/wiki/Solovay%E2%80%93Strassen_primality_test
    """

    # calculates the upper range of values that are going
    # to be used for the generation of numbers
    q = 0.5
    t = ceil_integer(k / math.log(1 / q, 2))

    # iterates over the range of t value plus one this is the
    # range that is going to be used for primality testing
    for _index in range(t + 1):

        # generates a random number in the interval and verifies
        # if the number is a jacobi witness to the number that
        # is going to be verified
        random_number = random_integer_interval(1, number - 1)
        is_witness = jacobi_witness(random_number, number)
        if is_witness: return False

    # returns valid as no jacobi witness has been found
    # for the current number that is being verified
    return True

def jacobi_witness(x, n):
    """
    Checks if the given x value is witness to n value
    non primality.
    This check is made according to euler's theorem.
    The provided value x is considered to be a witness
    to n in case it is an euler's pseudo-prime with base x

    :type x: int
    :param x: The value to be checked for witness.
    :type n: int
    :param n: The value to be checked for primality.
    :rtype: bool
    :return: The result of the checking, if it passed
    the test or not (is witness or not).
    """

    j = jacobi(x, n) % n
    f = pow(x, (n - 1) // 2, n)

    if j == f: return False
    else: return True

def jacobi(a, b):
    """
    Calculates the value of the jacobi symbol, using the
    given a and b values.

    The possible return values for jacobi symbols are:
    -1, 0 or 1.

    :type a: int
    :param a: The a value of the jacobi symbol.
    :type b: int
    :param b: The b value of the jacobi symbol.
    :rtype: int
    :return: The calculated jacobi symbol, possible values
    are: -1, 0 or 1.
    :see: http://en.wikipedia.org/wiki/Jacobi_symbol
    """

    if a % b == 0: return 0

    result = 1

    while a > 1:
        if a & 1:
            if ((a - 1) * (b - 1) >> 2) & 1: result = -result
            b, a = a, b % a
        else:
            if ((b ** 2 - 1) >> 3) & 1: result = -result
            a >>= 1

    return result

def ceil_integer(value):
    """
    Retrieves the ceil of a value and then converts it
    into a valid integer for integer computation.
    The conversion to integer ensures that the ceil
    is compatible with certain (integer) operations.

    :type value: int
    :param value: The value to apply the ceil and that
    will latter be converted into a valid integer.
    :rtype: int
    :return: The ceil of the given value "casted" as an
    integer to be able to be used in integer math.
    """

    value = math.ceil(value)
    value = int(value)
    return value



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import os
import re
import datetime
import mimetypes

import netius.common
import netius.servers

BUFFER_SIZE = 32768
""" The size of the buffer that is going to be used when
sending the file to the client, this should not be neither
to big nor to small (as both situations would create problems) """

FOLDER_SVG = "<svg aria-hidden=\"true\" class=\"octicon octicon-file-directory\" height=\"16\" version=\"1.1\" viewBox=\"0 0 14 16\" width=\"14\"><path d=\"M13 4H7V3c0-.66-.31-1-1-1H1c-.55 0-1 .45-1 1v10c0 .55.45 1 1 1h12c.55 0 1-.45 1-1V5c0-.55-.45-1-1-1zM6 4H1V3h5v1z\"></path></svg>"
""" The vector code to be used for the icon that represents
a folder under the directory listing """

FILE_SVG = "<svg aria-hidden=\"true\" class=\"octicon octicon-file-text\" height=\"16\" version=\"1.1\" viewBox=\"0 0 12 16\" width=\"12\"><path d=\"M6 5H2V4h4v1zM2 8h7V7H2v1zm0 2h7V9H2v1zm0 2h7v-1H2v1zm10-7.5V14c0 .55-.45 1-1 1H1c-.55 0-1-.45-1-1V2c0-.55.45-1 1-1h7.5L12 4.5zM11 5L8 2H1v12h10V5z\"></path></svg>"
""" The vector code to be used for the icon that represents
a plain file under the directory listing """

EMPTY_GIF = "data:image/gif;base64,R0lGODlhAQABAAAAACH5BAEKAAEALAAAAAABAAEAAAICTAEAOw=="
""" Simple base 64 encoded empty gif to avoid possible image
corruption while rendering empty images on browser """

class FileServer(netius.servers.HTTP2Server):
    """
    Simple implementation of a file server that is able to list files
    for directories taking into account the base path values.

    This is a synchronous implementation meaning that the server loop
    will block for the various I/O operations to be performed.

    Current implementation supports byte ranges so that partial retrieval
    of a file is possible.
    """

    def __init__(
        self,
        base_path = "",
        style_urls = [],
        index_files = [],
        path_regex = [],
        list_dirs = True,
        list_engine = "base",
        cors = False,
        cache = 0,
        *args,
        **kwargs
    ):
        netius.servers.HTTP2Server.__init__(self, *args, **kwargs)
        self.base_path = base_path
        self.style_urls = style_urls
        self.index_files = index_files
        self.path_regex = path_regex
        self.list_dirs = list_dirs
        self.list_engine = list_engine
        self.cors = cors
        self.cache = cache

    @classmethod
    def _sorter_build(cls, name = None):

        def sorter(item):
            is_dir = item["is_dir"]
            is_top = item["name"] == ".."
            is_dir_v = 0 if is_dir else 1
            is_dir_v = -1 if is_top else is_dir_v

            if name == "name": return (item["name"], is_dir_v)
            if name == "modified": return (item["modified"], is_dir_v)
            if name == "size": return (item["size"], is_dir_v)
            if name == "type": return (item["type"], is_dir_v)

            return (is_dir_v, item["name"])

        return sorter

    @classmethod
    def _items_normalize(
        cls,
        items,
        path,
        pad = False,
        space = True,
        simplified = False
    ):
        _items = []

        for item in items:
            if netius.legacy.PYTHON_3: item_s = item
            else: item_s = item.encode("utf-8")

            path_f = os.path.join(path, item)
            if not os.path.exists(path_f): continue

            is_dir = os.path.isdir(path_f)
            item_s = item_s + "/" if is_dir and pad else item_s
            item_q = netius.legacy.quote(item_s)

            _time = os.path.getmtime(path_f)
            date_time = datetime.datetime.utcfromtimestamp(_time)
            time_s = date_time.strftime("%Y-%m-%d %H:%M")

            size = 0 if is_dir else os.path.getsize(path_f)
            size_s = netius.common.size_round_unit(
                size,
                space = space,
                simplified = simplified
            )
            size_s = "-" if is_dir else size_s

            type_s, _encoding = mimetypes.guess_type(path_f, strict = True)
            type_s = type_s or "-"
            type_s = "Directory" if is_dir else type_s

            icon = FOLDER_SVG if is_dir else FILE_SVG

            _item = dict(
                name = item,
                name_s = item_s,
                name_q = item_q,
                is_dir = is_dir,
                path = path_f,
                modified = time_s,
                size = size,
                size_s = size_s,
                type = type_s,
                type_s = type_s,
                icon = icon
            )

            _items.append(_item)

        return _items

    @classmethod
    def _gen_dir(cls, engine, path, path_v, query_m, style = True, style_urls = [], **kwargs):
        gen_dir_method = getattr(cls, "_gen_dir_" + engine)
        return gen_dir_method(
            path,
            path_v,
            query_m,
            style = style,
            style_urls = style_urls,
            **kwargs
        )

    @classmethod
    def _gen_dir_base(cls, path, path_v, query_m, style = True, style_urls = [], **kwargs):
        sort = query_m.get("sort", [])
        direction = query_m.get("direction", [])

        sort = sort[0] if sort else None
        direction = direction[0] if direction else "asc"

        reverse = direction == "desc"
        _direction = "desc" if direction == "asc" else "asc"

        items = os.listdir(path)

        is_root = path_v == "" or path_v == "/"
        if not is_root: items.insert(0, "..")

        items = cls._items_normalize(items, path, pad = not style)
        items.sort(key = lambda v: v["name"])
        items.sort(
            key = cls._sorter_build(name = sort),
            reverse = reverse
        )

        path_n = path_v.rstrip("/")

        path_b = []
        current = str()
        paths = path_n.split("/")

        for item in paths[:-1]:
            current += item + "/"
            path_b.append(" <a href=\"%s\">%s</a> " % (current, item or "/"))
            if not item: continue
            path_b.append("<span>/</span>")

        path_b.append(" <span>%s</span>" % (paths[-1] or "/"))
        path_s = "".join(path_b)
        path_s = path_s.strip()

        for value in cls._gen_header(
            "Index of %s" % (path_n or "/"),
            style = style,
            style_urls = style_urls
        ):
            yield value

        yield "<body>"
        yield "<h1 class=\"path\">Index of %s</h1>" % path_s
        yield "<hr/>"
        yield "<table>"
        yield "<thead>"
        yield "<tr>"
        yield "<th align=\"left\" width=\"350\">"
        yield "<a href=\"?sort=name&direction=%s\" class=\"%s\">Name</a>" %\
            (_direction, "selected" if sort == "name" else "")
        yield "</th>"
        yield "<th align=\"left\" width=\"130\">"
        yield "<a href=\"?sort=modified&direction=%s\" class=\"%s\">Last Modified</a>" %\
            (_direction, "selected" if sort == "modified" else "")
        yield "</th>"
        yield "<th align=\"left\" width=\"80\">"
        yield "<a href=\"?sort=size&direction=%s\" class=\"%s\">Size</a></th>" %\
            (_direction, "selected" if sort == "size" else "")
        yield "</th>"
        yield "<th align=\"left\" width=\"200\">"
        yield "<a href=\"?sort=type&direction=%s\" class=\"%s\">Type</a></th>" %\
            (_direction, "selected" if sort == "type" else "")
        yield "</th>"
        yield "</tr>"
        yield "</thead>"
        yield "<tbody>"
        for item in items:
            yield "<tr>"
            yield "<td>"
            if style: yield item["icon"]
            yield "<a href=\"%s\">%s</a>" % (item["name_q"], item["name_s"])
            yield "</td>"
            yield "<td>%s</td>" % item["modified"]
            yield "<td>%s</td>" % item["size_s"]
            yield "<td>%s</td>" % item["type_s"]
            yield "</tr>"
        yield "</tbody>"
        yield "</table>"
        yield "<hr/>"
        yield "<span>"
        yield netius.IDENTIFIER
        yield "</span>"
        yield "</body>"

        for value in cls._gen_footer(): yield value

    @classmethod
    def _gen_dir_apache(cls, path, path_v, query_m, **kwargs):
        sort = query_m.get("sort", [])
        direction = query_m.get("direction", [])

        sort = sort[0] if sort else None
        direction = direction[0] if direction else "asc"

        reverse = direction == "desc"
        _direction = "desc" if direction == "asc" else "asc"

        items = os.listdir(path)

        items.insert(0, "..")

        items = cls._items_normalize(
            items,
            path,
            pad = True,
            space = False,
            simplified = True
        )
        items.sort(key = lambda v: v["name"])
        items.sort(
            key = cls._sorter_build(name = sort),
            reverse = reverse
        )

        path_n = path_v.rstrip("/")

        for value in cls._gen_header("Index of %s" % (path_n or "/"), style = False, meta = False):
            yield value

        yield "<body>"
        yield "<h1 class=\"path\">Index of %s</h1>" % (path_n or "/")
        yield "<table>"
        yield "<tr>"
        yield "<th valign=\"top\"><img src=\"%s\" alt=\"[ICO]\"></th>" % EMPTY_GIF
        yield "<th>"
        yield "<a href=\"?sort=name&direction=%s\" class=\"%s\">Name</a>" %\
            (_direction, "selected" if sort == "name" else "")
        yield "</th>"
        yield "<th>"
        yield "<a href=\"?sort=modified&direction=%s\" class=\"%s\">Last modified</a>" %\
            (_direction, "selected" if sort == "modified" else "")
        yield "</th>"
        yield "<th>"
        yield "<a href=\"?sort=size&direction=%s\" class=\"%s\">Size</a></th>" %\
            (_direction, "selected" if sort == "size" else "")
        yield "</th>"
        yield "<th>"
        yield "<a href=\"?sort=description&direction=%s\" class=\"%s\">Description</a></th>" %\
            (_direction, "selected" if sort == "description" else "")
        yield "</th>"
        yield "</tr>"
        yield "<tr><th colspan=\"5\"><hr></th></tr>"
        for item in items:
            if item["name_s"] == "..": type_s = "PARENTDIR"
            elif item["is_dir"]: type_s = "DIR"
            else: type_s = "ARC"
            if item["name_s"] == "..": name_s = "Parent Directory"
            elif item["is_dir"]: name_s = item["name_s"] + "/"
            else: name_s = item["name_s"]
            if item["is_dir"]: name_q = item["name_q"] + "/"
            else: name_q = item["name_q"]
            yield "<tr>"
            yield "<td valign=\"top\"><img src=\"%s\" alt=\"[%s]\"></td>" % (EMPTY_GIF, type_s)
            yield "<td><a href=\"%s\">%s</a></td>" % (name_q, name_s)
            yield "<td>%s</td>" % item["modified"]
            yield "<td align=\"right\">%s</td>" % item["size_s"]
            yield "<td>%s</td>" % item["type_s"]
            yield "</tr>"
            yield "\n"
        yield "<tr><th colspan=\"5\"><hr></th></tr>"
        yield "</table>"
        yield "<address>%s</address>" % netius.IDENTIFIER
        yield "</body>"

        for value in cls._gen_footer(): yield value

    @classmethod
    def _gen_dir_legacy(cls, path, path_v, query_m, **kwargs):
        max_length = kwargs.get("max_length", 24)
        spacing = kwargs.get("spacing", 2)

        sort = query_m.get("sort", [])
        direction = query_m.get("direction", [])

        sort = sort[0] if sort else None
        direction = direction[0] if direction else "asc"

        reverse = direction == "desc"
        _direction = "desc" if direction == "asc" else "asc"

        items = os.listdir(path)

        items.insert(0, "..")

        items = cls._items_normalize(
            items,
            path,
            pad = True,
            space = False,
            simplified = True
        )
        items.sort(key = lambda v: v["name"])
        items.sort(
            key = cls._sorter_build(name = sort),
            reverse = reverse
        )

        max_length = max([len(item["name_s"]) for item in items] + [max_length])
        padding_s = (max_length + spacing - 4) * " "
        spacing_s = spacing * " "

        path_n = path_v.rstrip("/")

        for value in cls._gen_header("Index of %s" % (path_n or "/"), style = False, meta = False):
            yield value

        yield "<body>"
        yield "<h1 class=\"path\">Index of %s</h1>" % (path_n or "/")
        yield "<hr/>"
        yield "<pre>"
        yield "<img src=\"%s\" alt=\"Icon \">" % EMPTY_GIF
        yield "<a href=\"?sort=name&direction=%s\" class=\"%s\">Name</a>" %\
            (_direction, "selected" if sort == "name" else "")
        yield padding_s
        yield "<a href=\"?sort=modified&direction=%s\" class=\"%s\">Last modified</a>" %\
            (_direction, "selected" if sort == "modified" else "")
        yield "   "
        yield spacing_s
        yield "<a href=\"?sort=size&direction=%s\" class=\"%s\">Size</a></th>" %\
            (_direction, "selected" if sort == "size" else "")
        yield "<hr/>"
        for item in items:
            if item["name_s"] == "..": type_s = "PARENTDIR"
            elif item["is_dir"]: type_s = "DIR"
            else: type_s = "ARC"
            if item["name_s"] == "..": name_s = "Parent Directory"
            elif item["is_dir"]: name_s = item["name_s"] + "/"
            else: name_s = item["name_s"]
            if item["is_dir"]: name_q = item["name_q"] + "/"
            else: name_q = item["name_q"]
            name_s = name_s[:max_length]
            padding_r = max_length - len(name_s)
            yield "<img src=\"%s\" alt=\"[%s]\" />" % (EMPTY_GIF, type_s)
            yield "<a href=\"%s\">%s</a>" % (name_q, name_s)
            yield " " * padding_r
            yield spacing_s
            yield "%s%s%s" % (item["modified"], spacing_s, item["size_s"].ljust(5))
            yield spacing_s
            yield "\n"
        yield "<hr/>"
        yield "</pre>"
        yield "<address>%s</address>" % netius.IDENTIFIER
        yield "</body>"

        for value in cls._gen_footer(): yield value

    def on_connection_d(self, connection):
        netius.servers.HTTP2Server.on_connection_d(self, connection)

        file = hasattr(connection, "file") and connection.file
        if file: file.close()
        setattr(connection, "file", None)
        setattr(connection, "range", None)
        setattr(connection, "bytes_p", None)
        setattr(connection, "queue", None)

    def on_stream_d(self, stream):
        file = hasattr(stream, "file") and stream.file
        if file: file.close()
        setattr(stream, "file", None)
        setattr(stream, "range", None)
        setattr(stream, "bytes_p", None)
        setattr(stream, "queue", None)

    def on_serve(self):
        netius.servers.HTTP2Server.on_serve(self)
        if self.env: self.base_path = self.get_env("BASE_PATH", self.base_path)
        if self.env: self.style_urls = self.get_env("STYLE_URLS", self.style_urls, cast = list)
        if self.env: self.index_files = self.get_env("INDEX_FILES", self.index_files, cast = list)
        if self.env: self.path_regex = self.get_env(
            "PATH_REGEX",
            self.path_regex,
            cast = lambda v: [i.split(":") for i in v.split(";")]
        )
        if self.env: self.list_dirs = self.get_env("LIST_DIRS", self.list_dirs, cast = bool)
        if self.env: self.list_engine = self.get_env("LIST_ENGINE", self.list_engine)
        if self.env: self.cors = self.get_env("CORS", self.cors, cast = bool)
        if self.env: self.cache = self.get_env("CACHE", self.cache, cast = int)
        self._build_regex()
        self.base_path = os.path.abspath(self.base_path)
        self.cache_d = datetime.timedelta(seconds = self.cache)
        self.base_path = netius.legacy.u(self.base_path, force = True)
        self.info("Defining '%s' as the root of the file server ..." % (self.base_path or "."))
        if self.list_dirs: self.info("Listing directories with '%s' engine ..." % self.list_engine)
        if self.cors: self.info("Cross origin resource sharing is enabled")
        if self.cache: self.info("Resource cache set with %d seconds" % self.cache)

    def on_data_http(self, connection, parser):
        netius.servers.HTTP2Server.on_data_http(self, connection, parser)

        # verifies if the current connection contains a reference to the
        # file object, in case it exists there's a file currently being
        # handled by the connection and so the current data processing
        # must be delayed until the file is processed (inserted in queue)
        if hasattr(connection, "file") and connection.file:
            if not hasattr(connection, "queue"): connection.queue = []
            state = parser.get_state()
            connection.queue.append(state)
            return

        try:
            # retrieves the requested path from the parser and the constructs
            # the correct file name/path to be used in the reading from the
            # current file system, so that it's possible to handle the data
            path = parser.get_path(normalize = True)
            path = netius.legacy.unquote(path)
            path = path.lstrip("/")
            path = netius.legacy.u(path, force = True)
            path = self._resolve(path)
            path_f = os.path.join(self.base_path, path)
            path_f = os.path.abspath(path_f)
            path_f = os.path.normpath(path_f)

            # retrieves the current file system encoding and determines if it
            # it's required to decode the path into an unicode string, if that's
            # the case the normal decoding process is used using the currently
            # defined file system encoding as defined in the specification
            path_f = netius.legacy.u(path_f, encoding = "utf-8", force = True)

            # verifies if the provided path starts with the contents of the
            # base path in case it does not it's a security issue and a proper
            # exception must be raised indicating the issue
            is_sub = path_f.startswith(self.base_path)
            if not is_sub: raise netius.SecurityError("Invalid path")

            # verifies if the requested file exists in case it does not
            # raises an error indicating the problem so that the user is
            # notified about the failure to find the appropriate file
            if not os.path.exists(path_f): self.on_no_file(connection); return

            # verifies if the currently resolved path refers an directory or
            # instead a normal file and handles each of the cases properly by
            # redirecting the request to the proper handlers
            is_dir = os.path.isdir(path_f)
            if is_dir: self.on_dir_file(connection, parser, path_f)
            else: self.on_normal_file(connection, parser, path_f)
        except BaseException as exception:
            # handles the exception gracefully by sending the contents of
            # it to the client and identifying the problem correctly
            self.on_exception_file(connection, exception)

    def on_dir_file(self, connection, parser, path, style = True):
        cls = self.__class__

        path_v = parser.get_path()
        path_v = netius.legacy.unquote(path_v)

        query_v = parser.get_query()
        query_m = parser._parse_query(query_v)

        is_valid = path_v.endswith("/")
        if not is_valid:
            path_q = netius.legacy.quote(path_v)
            connection.send_response(
                data = "Permanent redirect",
                headers = dict(
                    location = path_q + "/"
                ),
                code = 301,
                apply = True
            )
            return

        for index_file in self.index_files:
            index_path = os.path.join(path, index_file)
            if not os.path.exists(index_path): continue
            return self.on_normal_file(connection, parser, index_path)

        if not self.list_dirs:
            self.on_no_file(connection)
            return

        data = "".join(cls._gen_dir(
            self.list_engine,
            path,
            path_v,
            query_m,
            style = style,
            style_urls = self.style_urls
        ))
        data = netius.legacy.bytes(data, encoding = "utf-8", force = True)

        headers = dict()
        headers["content-type"] = "text/html"

        connection.send_response(
            data = data,
            headers = headers,
            code = 200,
            apply = True,
            callback = self._file_check_close
        )

    def on_normal_file(self, connection, parser, path):
        # encodes the current path in case it's currently represented by
        # a string, this is going to avoid problems in the logging of the
        # path that is being requested (unicode encoding problems)
        path_s = path if netius.legacy.is_str(path) else path.encode("utf-8")

        # prints a debug message about the file that is going to be read
        # from the current file system to be sent to the connection
        self.debug("Reading file '%s' from file system" % path_s)

        # uses the parser from the connection to be able to gather the
        # range as a string to be used latter for conversion
        range_s = parser.headers.get("range", None)
        is_partial = True if range_s else False

        # retrieves the last modified timestamp for the resource path and
        # uses it to create the ETag for the resource to be served
        modified = os.path.getmtime(path)
        etag = "netius-%.2f" % modified

        # retrieves the header that describes the previous version in the
        # client side (client side ETag) and compares both of the ETags to
        # verify if the file changed meanwhile or not
        _etag = parser.headers.get("if-none-match", None)
        not_modified = etag == _etag

        # in case the file did not change in the mean time the not modified
        # callback must be called to correctly handled the file no change
        if not_modified: self.on_not_modified(connection, path); return

        # tries to guess the mime type of the file present in the target
        # file path that is going to be returned, this may fail as it's not
        # always possible to determine the correct mime type for a file
        # for suck situations the default mime type is used
        type, _encoding = mimetypes.guess_type(path, strict = True)
        type = type or "application/octet-stream"

        # retrieves the size of the file that has just be resolved using
        # the currently provided path value and then associates the file
        # with the current connection
        file_size = os.path.getsize(path)
        file = open(path, "rb")
        connection.file = file

        # convert the current string based representation of the range
        # into a tuple based presentation otherwise creates the default
        # tuple containing the initial position and the final one
        if is_partial:
            range_s = range_s[6:]
            start_s, end_s = range_s.split("-", 1)
            start = int(start_s) if start_s else 0
            end = int(end_s) if end_s else file_size - 1
            range = (start, end)
        else: range = (0, file_size - 1)

        # calculates the real data size of the chunk that is going to be
        # sent to the client this must use the normal range approach
        data_size = range[1] - range[0] + 1

        # associates the range tuple with the current connection as it's
        # going to be used latter for additional computation
        connection.range = range
        connection.bytes_p = data_size

        # seeks the current file to the initial position where it's going
        # to start it's reading processing as according to the range
        file.seek(range[0])

        # creates the string that will represent the content range that is
        # going to be returned to the client in the current request
        content_range_s = "bytes %d-%d/%d" % (range[0], range[1], file_size)

        # creates the map that will hold the various header values for the
        # the current message to be sent it may contain both the length
        # of the file that is going to be returned and the type of it
        headers = dict()
        headers["etag"] = etag
        headers["content-length"] = "%d" % data_size
        if self.cors: headers["access-control-allow-origin"] = "*"
        if type: headers["content-type"] = type
        if is_partial: headers["content-range"] = content_range_s
        if not is_partial: headers["accept-ranges"] = "bytes"

        # in case there's a valid cache defined must populate the proper header
        # fields so that cache is applied to the request
        if self.cache:
            current = datetime.datetime.utcnow()
            target = current + self.cache_d
            target_s = target.strftime("%a, %d %b %Y %H:%M:%S GMT")
            cache_s = "public, max-age=%d" % self.cache
            headers["expires"] = target_s
            headers["cache-control"] = cache_s

        # "calculates" the proper returning code taking into account if the
        # current data to be sent is partial or not
        code = 206 if is_partial else 200

        # sends the initial part of the file response containing the headers
        # and the description of the file (includes size) the callback to this
        # operation is the initial sending of the file contents so that the
        # sending of the proper file contents starts with success
        connection.send_response(
            headers = headers,
            code = code,
            apply = True,
            final = False,
            flush = False,
            callback = self._file_send
        )

    def on_no_file(self, connection):
        cls = self.__class__
        connection.send_response(
            data = cls.build_text(
                "File not found",
                style_urls = self.style_urls
            ),
            headers = dict(
                connection = "close"
            ),
            code = 404,
            apply = True,
            callback = self._file_close
        )

    def on_exception_file(self, connection, exception):
        cls = self.__class__
        connection.send_response(
            data = cls.build_text(
                "Problem handling request - %s" % str(exception),
                trace = self.is_devel(),
                style_urls = self.style_urls
            ),
            headers = dict(
                connection = "close"
            ),
            code = 500,
            apply = True,
            callback = self._file_close
        )

    def on_not_modified(self, connection, path):
        connection.set_encoding(netius.common.PLAIN_ENCODING)
        connection.send_response(
            data = "",
            code = 304,
            apply = True
        )

    def _next_queue(self, connection):
        # verifies if the current connection already contains a reference to
        # the queue structure that handles the queuing/pipelining of requests
        # if it does not or the queue is empty returns immediately, as there's
        # nothing currently pending to be done/processed
        if not hasattr(connection, "queue"): return
        if not connection.queue: return

        # retrieves the state (of the parser) as the payload of the next element
        # in the queue and then uses it to construct a mock parser object that is
        # going to be used to simulate an on data call to the file server
        state = connection.queue.pop(0)
        parser = netius.common.HTTPParser.mock(connection.parser.owner, state)
        try: self.on_data_http(connection, parser)
        finally: parser.destroy()

    def _file_send(self, connection):
        file = connection.file
        range = connection.range
        is_larger = BUFFER_SIZE > connection.bytes_p
        buffer_s = connection.bytes_p if is_larger else BUFFER_SIZE
        data = file.read(buffer_s)
        data_l = len(data) if data else 0
        connection.bytes_p -= data_l
        is_final = not data or connection.bytes_p == 0
        callback = self._file_finish if is_final else self._file_send
        connection.send_part(
            data,
            final = False,
            callback = callback
        )

    def _file_finish(self, connection):
        connection.file.close()
        connection.file = None
        connection.range = None
        connection.bytes_p = None
        is_keep_alive = connection.parser.keep_alive
        callback = None if is_keep_alive else self._file_close
        connection.flush_s(callback = callback)
        self._next_queue(connection)

    def _file_close(self, connection):
        connection.close(flush = True)

    def _file_check_close(self, connection):
        if connection.parser.keep_alive: return
        connection.close(flush = True)

    def _resolve(self, path):
        path, result = self._resolve_regex(path)
        if result: return path
        return path

    def _build_regex(self):
        self.path_regex = [(re.compile(regex), value) for regex, value in self.path_regex]

    def _resolve_regex(self, path):
        for regex, value in self.path_regex:
            if not regex.match(path): continue
            return (value, True)
        return (path, False)

if __name__ == "__main__":
    import logging

    server = FileServer(level = logging.INFO)
    server.serve(env = True)
else:
    __path__ = []



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

from . import address
from . import allow
from . import base
from . import deny
from . import dummy
from . import memory
from . import passwd
from . import simple

from .address import AddressAuth
from .allow import AllowAuth
from .base import Auth
from .deny import DenyAuth
from .dummy import DummyAuth
from .memory import MemoryAuth
from .passwd import PasswdAuth
from .simple import SimpleAuth



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import base64
import hashlib

import netius.common

class WSConnection(netius.Connection):
    """
    Connection based class for the websockets connection,
    should be able to implement the required encoding
    and decoding techniques in compliance with the websockets
    level 13 specification.

    :see: http://tools.ietf.org/html/rfc6455
    """

    def __init__(self, *args, **kwargs):
        netius.Connection.__init__(self, *args, **kwargs)
        self.handshake = False
        self.method = None
        self.path = None
        self.version = None
        self.buffer_l = []
        self.headers = {}

    def send_ws(self, data):
        encoded = netius.common.encode_ws(data, mask = False)
        return self.send(encoded)

    def recv_ws(self, size = netius.CHUNK_SIZE):
        data = self.recv(size = size)
        decoded = netius.common.decode_ws(data)
        return decoded

    def add_buffer(self, data):
        self.buffer_l.append(data)

    def get_buffer(self, delete = True):
        if not self.buffer_l: return b""
        buffer = b"".join(self.buffer_l)
        if delete: del self.buffer_l[:]
        return buffer

    def do_handshake(self):
        if self.handshake:
            raise netius.NetiusError("Handshake already done")

        buffer = b"".join(self.buffer_l)
        end_index = buffer.find(b"\r\n\r\n")
        if end_index == -1: raise netius.DataError("Missing data for handshake")

        data = buffer[:end_index + 4]
        remaining = buffer[end_index + 4:]

        lines = data.split(b"\r\n")
        for line in lines[1:]:
            values = line.split(b":", 1)
            values_l = len(values)
            if not values_l == 2: continue

            key, value = values
            key = key.strip()
            key = netius.legacy.str(key)
            value = value.strip()
            value = netius.legacy.str(value)
            self.headers[key] = value

        first = lines[0]
        first = netius.legacy.str(first)
        self.method, self.path, self.version = first.split(" ", 2)

        del self.buffer_l[:]
        self.handshake = True

        if remaining: self.add_buffer(remaining)

    def accept_key(self):
        socket_key = self.headers.get("Sec-WebSocket-Key", None)
        if not socket_key:
            raise netius.NetiusError("No socket key found in headers")

        value = netius.legacy.bytes(socket_key + WSServer.MAGIC_VALUE)
        hash = hashlib.sha1(value)
        hash_digest = hash.digest()
        accept_key = base64.b64encode(hash_digest)
        accept_key = netius.legacy.str(accept_key)
        return accept_key

class WSServer(netius.StreamServer):
    """
    Base class for the creation of websocket server, should
    handle both the upgrading/handshaking of the connection
    and together with the associated connection class the
    encoding and decoding of the frames.
    """

    MAGIC_VALUE = "258EAFA5-E914-47DA-95CA-C5AB0DC85B11"
    """ The magic value used by the websocket protocol as part
    of the key generation process in the handshake """

    def on_data(self, connection, data):
        netius.StreamServer.on_data(self, connection, data)

        # iterates while there's still data pending to be parsed from the
        # current message received using the http or ws protocols
        while data:
            if connection.handshake:
                # retrieves the current (pending) buffer of data for the
                # connection and tries to run the decoder of websockets
                # frame on the complete set of data pending in case there's
                # a problem the (pending) data is added to the buffer
                buffer = connection.get_buffer()
                data = buffer + data
                try: decoded, data = netius.common.decode_ws(data)
                except netius.DataError: connection.add_buffer(data); break
                self.on_data_ws(connection, decoded)

            else:
                # adds the current data to the internal connection
                # buffer to be processed latter, by the handshake
                connection.add_buffer(data)

                # tries to run the handshake operation for the
                # current connection in case it fails due to an
                # handshake error must delay the execution to the
                # next iteration (not enough data)
                try: connection.do_handshake()
                except netius.DataError: return

                # retrieves (and computes) the accept key value for
                # the current request and sends it as the handshake
                # response to the client side
                accept_key = connection.accept_key()
                response = self._handshake_response(accept_key)
                connection.send(response)

                # calls the on handshake event handler for the current
                # connection to notify the current object
                self.on_handshake(connection)

                # retrieves the current buffer value as the data that is still
                # pending to be parsed from the current connection, this is
                # required so that the complete client buffer is flushed
                data = connection.get_buffer()

    def build_connection(self, socket, address, ssl = False):
        return WSConnection(
            owner = self,
            socket = socket,
            address = address,
            ssl = ssl
        )

    def send_ws(self, connection, data):
        encoded = netius.common.encode_ws(data, mask = False)
        return connection.send(encoded)

    def on_data_ws(self, connection, data):
        pass

    def on_handshake(self, connection):
        pass

    def _handshake_response(self, accept_key):
        """
        Returns the response contents of the handshake operation for
        the provided accept key.

        The key value should already be calculated according to the
        specification.

        :type accept_key: String
        :param accept_key: The accept key to be used in the creation
        of the response message.
        :rtype: String
        :return: The response message contents generated according to
        the specification and the provided accept key.
        """

        data = "HTTP/1.1 101 Switching Protocols\r\n" +\
            "Upgrade: websocket\r\n" +\
            "Connection: Upgrade\r\n" +\
            "Sec-WebSocket-Accept: %s\r\n\r\n" % accept_key
        return data



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import struct

import netius

from . import util
from . import parser

IPV4 = 0x01

IPV6 = 0x04

DOMAIN = 0x03

VERSION_STATE = 1

HEADER_STATE = 2

USER_ID_STATE = 3

DOMAIN_STATE = 4

AUTH_COUNT_STATE = 5

AUTH_METHODS_STATE = 6

HEADER_EXTRA_STATE = 7

SIZE_STATE = 8

ADDRESS_STATE = 9

PORT_STATE = 10

FINISH_STATE = 11

class SOCKSParser(parser.Parser):

    def __init__(self, owner):
        parser.Parser.__init__(self, owner)

        self.build()
        self.reset()

    def build(self):
        """
        Builds the initial set of states ordered according to
        their internal integer definitions, this method provides
        a fast and scalable way of parsing data.
        """

        parser.Parser.build(self)

        self.states = (
            self._parse_version,
            self._parse_header,
            self._parse_user_id,
            self._parse_domain,
            self._parse_auth_count,
            self._parse_auth_methods,
            self._parse_header_extra,
            self._parse_size,
            self._parse_address,
            self._parse_port
        )
        self.state_l = len(self.states)

    def destroy(self):
        """
        Destroys the current structure for the parser meaning that
        it's restored to the original values, this method should only
        be called on situation where no more parser usage is required.
        """

        parser.Parser.destroy(self)

        self.states = ()
        self.state_l = 0

    def reset(self):
        self.state = VERSION_STATE
        self.buffer = []
        self.version = None
        self.command = None
        self.port = None
        self.address = None
        self.address_s = None
        self.user_id = None
        self.domain = None
        self.type = None
        self.size = 0
        self.is_extended = False
        self.auth_count = 0
        self.auth_methods = None

    def clear(self, force = False):
        if not force and self.state == VERSION_STATE: return
        self.reset()

    def parse(self, data):
        """
        Parses the provided data chunk, changing the current
        state of the parser accordingly and returning the
        number of processed bytes from it.

        :type data: String
        :param data: The string containing the data to be parsed
        in the current parse operation.
        :rtype: int
        :return: The amount of bytes of the data string that have
        been "parsed" in the current parse operation.
        """

        parser.Parser.parse(self, data)

        # in case the current state of the parser is finished, must
        # reset the state to the start position as the parser is
        # re-starting (probably a new data sequence)
        if self.state == FINISH_STATE: self.clear()

        # retrieves the size of the data that has been sent for parsing
        # and saves it under the size original variable
        size = len(data)
        size_o = size

        # iterates continuously to try to process all that
        # data that has been sent for processing
        while size > 0:

            if self.state <= self.state_l:
                method = self.states[self.state - 1]
                count = method(data)
                if count == 0: break

                size -= count
                data = data[count:]

                continue

            elif self.state == FINISH_STATE:
                break

            else:
                raise netius.ParserError("Invalid state '%d'" % self.state)

        # in case not all of the data has been processed
        # must add it to the buffer so that it may be used
        # latter in the next parsing of the message
        if size > 0: self.buffer.append(data)

        # returns the number of read (processed) bytes of the
        # data that has been sent to the parser
        return size_o - size

    def get_host(self):
        return self.domain or self.address_s

    def get_address(self):
        if self.type == None: return None

        if self.type == IPV4: address = struct.pack("!I", self.address)
        elif self.type == IPV6: address = struct.pack("!QQ", self.address)
        else: address = struct.pack("!B", self.size) + self.address

        return address

    def _parse_version(self, data):
        if len(data) < 1:
            raise netius.ParserError("Invalid request (too short)")

        request = data[:1]
        self.version, = struct.unpack("!B", request)

        if self.version == 4: self.state = HEADER_STATE
        elif self.version == 5: self.state = AUTH_COUNT_STATE
        else: raise netius.ParserError("Invalid version '%d'" % self.version)

        return 1

    def _parse_header(self, data):
        if len(data) < 7:
            raise netius.ParserError("Invalid request (too short)")

        request = data[:7]
        self.command, self.port, self.address = struct.unpack("!BHI", request)
        self.address_s = util.addr_to_ip4(self.address)

        self.is_extended = self.address_s.startswith("0.0.0.")

        self.state = USER_ID_STATE

        return 7

    def _parse_user_id(self, data):
        index = data.find(b"\0")
        if index == -1: return 0

        self.buffer.append(data[:index])
        self.user_id = b"".join(self.buffer)
        self.user_id = netius.legacy.str(self.user_id)
        del self.buffer[:]

        if self.is_extended: self.state = DOMAIN_STATE
        else: self.state = FINISH_STATE

        if not self.is_extended: self.trigger("on_data")
        return index + 1

    def _parse_domain(self, data):
        index = data.find(b"\0")
        if index == -1: return 0

        self.buffer.append(data[:index])
        self.domain = b"".join(self.buffer)
        self.domain = netius.legacy.str(self.domain)
        del self.buffer[:]

        self.state = FINISH_STATE

        self.trigger("on_data")
        return index + 1

    def _parse_auth_count(self, data):
        if len(data) < 1:
            raise netius.ParserError("Invalid request (too short)")

        request = data[:1]
        self.auth_count, = struct.unpack("!B", request)

        self.state = AUTH_METHODS_STATE

        return 1

    def _parse_auth_methods(self, data):
        is_ready = len(data) + len(self.buffer) >= self.auth_count
        if not is_ready: return 0

        remaining = self.auth_count - len(self.buffer)
        self.buffer.append(data[:remaining])
        data = b"".join(self.buffer)

        format = "!%dB" % self.auth_count
        self.auth_methods = struct.unpack(format, data)
        del self.buffer[:]

        self.state = HEADER_EXTRA_STATE

        self.trigger("on_auth")
        return remaining

    def _parse_header_extra(self, data):
        if len(data) < 4:
            raise netius.ParserError("Invalid request (too short)")

        request = data[:4]
        self.version, self.command, _reserved, self.type =\
            struct.unpack("!BBBB", request)

        if self.type == IPV4: self.size = 4
        elif self.type == IPV6: self.size = 16

        if self.type == DOMAIN: self.state = SIZE_STATE
        else: self.state = ADDRESS_STATE

        return 4

    def _parse_size(self, data):
        if len(data) < 1:
            raise netius.ParserError("Invalid request (too short)")

        request = data[:1]
        self.size, = struct.unpack("!B", request)

        self.state = ADDRESS_STATE

        return 1

    def _parse_address(self, data):
        is_ready = len(data) + len(self.buffer) >= self.size
        if not is_ready: return 0

        remaining = self.size - len(self.buffer)
        self.buffer.append(data[:remaining])
        data = b"".join(self.buffer)

        if self.type == IPV4:
            self.address, = struct.unpack("!I", data)
            self.address_s = util.addr_to_ip4(self.address)
        elif self.type == IPV6:
            address_t = struct.unpack("!QQ", data)
            self.address = (address_t[0] << 64) + address_t[1]
            self.address_s = util.addr_to_ip6(self.address)
        else:
            self.address = netius.legacy.str(data)
            self.address_s = netius.legacy.str(data)

        self.state = PORT_STATE

        return remaining

    def _parse_port(self, data):
        if len(data) < 2:
            raise netius.ParserError("Invalid request (too short)")

        request = data[:2]
        self.port, = struct.unpack("!H", request)

        self.state = FINISH_STATE

        self.trigger("on_data")
        return 2
